# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to build PHX-Variant from Honeybee Rooms"""

from typing import Dict, Set

from honeybee import room

from honeybee_ph import site, phi, phius
from honeybee_energy_ph.properties.load import equipment
from honeybee_energy_ph.hvac.ventilation import PhVentilationSystem
from honeybee_energy_ph.hvac.heating import PhHeatingSystem
from honeybee_energy_ph.hvac.cooling import PhCoolingSystem

from PHX.model import phx_site, project, ground, constructions, certification
from PHX.model.enums import phi_certification_phpp_9, phi_certification_phpp_10, phius_certification
from PHX.from_HBJSON import create_building, create_hvac, create_shw, create_elec_equip


def add_building_from_hb_room(_variant: project.PhxVariant,
                              _hb_room: room.Room,
                              _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
                              _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
                              group_components: bool = False) -> None:
    """Create the  PHX-Building with all Components and Zones based on a Honeybee-Room.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the building to.
        * _hb_room (room.Room): The honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.
        * group_components (bool): default=False. Set to true to have the converter
            group the components by assembly-type.

    Returns:
    --------
        * None
    """
    _variant.building.add_components(
        create_building.create_components_from_hb_room(_hb_room, _assembly_dict, _window_type_dict))
    _variant.building.add_zones(
        create_building.create_zones_from_hb_room(_hb_room))
    _variant.building.add_thermal_bridges(
        create_building.create_thermal_bridges_from_hb_room(_hb_room))

    if group_components:
        _variant.building.merge_opaque_components_by_assembly()
        _variant.building.merge_aperture_components_by_assembly()


def add_phius_certification_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Set all the PhxPhiusCertificationCriteria on a PhxVariant based on a Honeybee-Room's Building Segment.

    Arguments:
    ----------
        * _variant (project.Variant): The PhxVariant to set the PhxPhiusCertificationCriteria on.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    # alias cus' all this shit is deep in there...
    hbph_phius_cert: phius.PhiusCertification = _hb_room.properties.ph.ph_bldg_segment.phius_certification  # type: ignore # alias
    phx_phius_cert_criteria = _variant.phius_certification.phius_certification_criteria # alias
    phx_phius_cert_settings = _variant.phius_certification.phius_certification_settings # alias

    # some random bullshit
    phx_phius_cert_criteria.ph_certificate_criteria = hbph_phius_cert.certification_criteria
    phx_phius_cert_criteria.ph_selection_target_data = hbph_phius_cert.localization_selection_type
    
    # certification thresholds (for Phius, they change)
    phx_phius_cert_criteria.phius_annual_heating_demand = hbph_phius_cert.PHIUS2021_heating_demand
    phx_phius_cert_criteria.phius_annual_cooling_demand = hbph_phius_cert.PHIUS2021_cooling_demand
    phx_phius_cert_criteria.phius_peak_heating_load = hbph_phius_cert.PHIUS2021_heating_load
    phx_phius_cert_criteria.phius_peak_cooling_load = hbph_phius_cert.PHIUS2021_cooling_load

    # certification settings / types
    phx_phius_cert_settings.phius_building_category_type = phius_certification.PhiusCertificationBuildingCategoryType(
        hbph_phius_cert.building_category_type.number)
    phx_phius_cert_settings.phius_building_use_type = phius_certification.PhiusCertificationBuildingUseType(
        hbph_phius_cert.building_use_type.number)
    phx_phius_cert_settings.phius_building_status = phius_certification.PhiusCertificationBuildingStatus(
        hbph_phius_cert.building_status.number)
    phx_phius_cert_settings.phius_building_type = phius_certification.PhiusCertificationBuildingType(
        hbph_phius_cert.building_type.number)

    return None


def set_phx_phpp9_settings(_phx_settings: certification.PhxPhiCertificationSettings, _attributes:phi.PHPPSettings9) -> None:
    """Set the values of the PHX-PHI Settings object for PHPP-v9 Settings.
    
    Arguments:
    ----------
        * _phx_settings: certification.PhxPhiCertificationSettings
        * _hbph_settings:phi.PhiCertification)
    
    Returns:
    --------
        * None
    """
    _phx_settings.phi_building_category_type = phi_certification_phpp_9.PhiCertBuildingCategoryType(_attributes.building_category_type.number)
    _phx_settings.phi_building_use_type = phi_certification_phpp_9.PhiCertBuildingUseType(_attributes.building_use_type.number)
    _phx_settings.phi_building_ihg_type = phi_certification_phpp_9.PhiCertIHGType(_attributes.ihg_type.number)
    _phx_settings.phi_building_occupancy_type = phi_certification_phpp_9.PhiCertOccupancyType(_attributes.occupancy_type.number)

    _phx_settings.phi_certification_type = phi_certification_phpp_9.PhiCertType(_attributes.certification_type.number)
    _phx_settings.phi_certification_class = phi_certification_phpp_9.PhiCertClass(_attributes.certification_class.number)
    _phx_settings.phi_pe_type = phi_certification_phpp_9.PhiCertificationPEType(_attributes.primary_energy_type.number)
    _phx_settings.phi_enerphit_type = phi_certification_phpp_9.PhiCertEnerPHitType(_attributes.enerphit_type.number)
    _phx_settings.phi_retrofit_type = phi_certification_phpp_9.PhiCertRetrofitType(_attributes.retrofit_type.number)


def set_phx_phpp10_settings(_phx_settings: certification.PhxPhiCertificationSettings, _attributes:phi.PHPPSettings10) -> None:
    """Set the values of the PHX-PHI Settings object for PHPP-v10 Settings.
    
    Arguments:
    ----------
        * _phx_settings: certification.PhxPhiCertificationSettings
        * _hbph_settings:phi.PhiCertification)
    
    Returns:
    --------
        * None
    """
    _phx_settings.phi_building_use_type = phi_certification_phpp_10.PhiCertBuildingUseType(_attributes.building_use_type.number)
    _phx_settings.phi_building_ihg_type = phi_certification_phpp_10.PhiCertIHGType(_attributes.ihg_type.number)

    _phx_settings.phi_certification_type = phi_certification_phpp_10.PhiCertType(_attributes.certification_type.number)
    _phx_settings.phi_certification_class = phi_certification_phpp_10.PhiCertClass(_attributes.certification_class.number)
    _phx_settings.phi_pe_type = phi_certification_phpp_10.PhiCertificationPEType(_attributes.primary_energy_type.number)
    _phx_settings.phi_retrofit_type = phi_certification_phpp_10.PhiCertRetrofitType(_attributes.retrofit_type.number)


def add_phi_certification_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Set all the PhxPhiCertificationCriteria on a PhxVariant based on a Honeybee-Room's Building Segment.

    Arguments:
    ----------
        * _variant (project.Variant): The PhxVariant to set the PhxPhiCertificationCriteria to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """
    # alias cus' all this shit is deep in there...
    hbph_settings: phi.PhiCertification = _hb_room.properties.ph.ph_bldg_segment.phi_certification  # type: ignore
    phx_settings = _variant.phi_certification.phi_certification_settings # type: ignore

    if hbph_settings.attributes.phpp_version == 10:
        set_phx_phpp10_settings(phx_settings, hbph_settings.attributes)
    elif hbph_settings.attributes.phpp_version == 9:
        set_phx_phpp9_settings(phx_settings, hbph_settings.attributes)
    else:
        msg = f"Error: Unknown PHPP Settings Version? Got: '{hbph_settings.attributes.phpp_version}'"
        raise Exception(msg)

    return None


def add_PhxPhBuildingData_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Create and add a PhxPhBuildingData with data from a Honeybee-Room to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the PH-Building to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """
    ph_bldg = _variant.phius_certification.ph_building_data # alias

    ph_bldg.num_of_units = _hb_room.properties.ph.ph_bldg_segment.num_dwelling_units # type: ignore
    ph_bldg.num_of_floors = _hb_room.properties.ph.ph_bldg_segment.num_floor_levels # type: ignore

    # TODO: Foundations. For now: set to None
    ph_bldg.add_foundation(ground.PhxFoundation())

    # Set the airtightness for Building
    ph_bldg.airtightness_q50 = _hb_room.properties.energy.infiltration.flow_per_exterior_area * 3600 # type: ignore

    # Air temp Setpoints
    ph_bldg.setpoints.winter = _hb_room.properties.ph.ph_bldg_segment.set_points.winter # type: ignore
    ph_bldg.setpoints.summer = _hb_room.properties.ph.ph_bldg_segment.set_points.summer # type: ignore

    return None


def add_climate_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Copy PHX-Climate info from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the climate data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    ud_site: site.Site = _hb_room.properties.ph.ph_bldg_segment.site # type: ignore

    # -- Location
    _variant.site.location.latitude = ud_site.location.latitude
    _variant.site.location.longitude = ud_site.location.longitude
    _variant.site.location.site_elevation = ud_site.location.site_elevation
    _variant.site.location.climate_zone = ud_site.location.climate_zone
    _variant.site.location.hours_from_UTC = ud_site.location.hours_from_UTC

    # -- Basics
    _variant.site.climate.daily_temp_swing = ud_site.climate.summer_daily_temperature_swing
    _variant.site.climate.avg_wind_speed = ud_site.climate.average_wind_speed

    # -- PHPP Stuff
    _variant.site.phpp_codes.country_code = ud_site.phpp_library_codes.country_code
    _variant.site.phpp_codes.region_code = ud_site.phpp_library_codes.region_code
    _variant.site.phpp_codes.dataset_name = ud_site.phpp_library_codes.dataset_name

    # -- Ground
    _variant.site.ground.ground_thermal_conductivity = ud_site.climate.ground.ground_thermal_conductivity
    _variant.site.ground.ground_heat_capacity = ud_site.climate.ground.ground_heat_capacity
    _variant.site.ground.ground_density = ud_site.climate.ground.ground_density
    _variant.site.ground.depth_groundwater = ud_site.climate.ground.depth_groundwater
    _variant.site.ground.flow_rate_groundwater = ud_site.climate.ground.flow_rate_groundwater

    # -- Monthly Values
    _variant.site.climate.temperature_air = ud_site.climate.monthly_temps.air_temps.values
    _variant.site.climate.temperature_dewpoint = ud_site.climate.monthly_temps.dewpoints.values
    _variant.site.climate.temperature_sky = ud_site.climate.monthly_temps.sky_temps.values

    _variant.site.climate.radiation_north =ud_site.climate.monthly_radiation.north.values
    _variant.site.climate.radiation_east =ud_site.climate.monthly_radiation.east.values
    _variant.site.climate.radiation_south =ud_site.climate.monthly_radiation.south.values
    _variant.site.climate.radiation_west =ud_site.climate.monthly_radiation.west.values
    _variant.site.climate.radiation_global =ud_site.climate.monthly_radiation.glob.values

    # -- Peak Load Values
    _variant.site.climate.peak_heating_1.temperature_air = ud_site.climate.peak_loads.heat_load_1.temp
    _variant.site.climate.peak_heating_1.radiation_north = ud_site.climate.peak_loads.heat_load_1.rad_north
    _variant.site.climate.peak_heating_1.radiation_east = ud_site.climate.peak_loads.heat_load_1.rad_east
    _variant.site.climate.peak_heating_1.radiation_south = ud_site.climate.peak_loads.heat_load_1.rad_south
    _variant.site.climate.peak_heating_1.radiation_west = ud_site.climate.peak_loads.heat_load_1.rad_west
    _variant.site.climate.peak_heating_1.radiation_global = ud_site.climate.peak_loads.heat_load_1.rad_global

    _variant.site.climate.peak_heating_2.temperature_air = ud_site.climate.peak_loads.heat_load_2.temp
    _variant.site.climate.peak_heating_2.radiation_north = ud_site.climate.peak_loads.heat_load_2.rad_north
    _variant.site.climate.peak_heating_2.radiation_east = ud_site.climate.peak_loads.heat_load_2.rad_east
    _variant.site.climate.peak_heating_2.radiation_south = ud_site.climate.peak_loads.heat_load_2.rad_south
    _variant.site.climate.peak_heating_2.radiation_west = ud_site.climate.peak_loads.heat_load_2.rad_west
    _variant.site.climate.peak_heating_2.radiation_global = ud_site.climate.peak_loads.heat_load_2.rad_global

    _variant.site.climate.peak_cooling_1.temperature_air = ud_site.climate.peak_loads.cooling_load_1.temp
    _variant.site.climate.peak_cooling_1.radiation_north = ud_site.climate.peak_loads.cooling_load_1.rad_north
    _variant.site.climate.peak_cooling_1.radiation_east = ud_site.climate.peak_loads.cooling_load_1.rad_east
    _variant.site.climate.peak_cooling_1.radiation_south = ud_site.climate.peak_loads.cooling_load_1.rad_south
    _variant.site.climate.peak_cooling_1.radiation_west = ud_site.climate.peak_loads.cooling_load_1.rad_west
    _variant.site.climate.peak_cooling_1.radiation_global = ud_site.climate.peak_loads.cooling_load_1.rad_global

    _variant.site.climate.peak_cooling_2.temperature_air = ud_site.climate.peak_loads.cooling_load_2.temp
    _variant.site.climate.peak_cooling_2.radiation_north = ud_site.climate.peak_loads.cooling_load_2.rad_north
    _variant.site.climate.peak_cooling_2.radiation_east = ud_site.climate.peak_loads.cooling_load_2.rad_east
    _variant.site.climate.peak_cooling_2.radiation_south = ud_site.climate.peak_loads.cooling_load_2.rad_south
    _variant.site.climate.peak_cooling_2.radiation_west = ud_site.climate.peak_loads.cooling_load_2.rad_west
    _variant.site.climate.peak_cooling_2.radiation_global = ud_site.climate.peak_loads.cooling_load_2.rad_global

    return None


def add_local_pe_conversion_factors(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Copy local Site->Source conversion factors from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the factor data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """
    for factor in _hb_room.properties.ph.ph_bldg_segment.source_energy_factors:# type: ignore
        new_phx_factor = phx_site.PhxPEFactor()
        new_phx_factor.fuel_name = factor.fuel_name
        new_phx_factor.value = factor.value
        new_phx_factor.unit = factor.unit
        _variant.site.energy_factors.pe_factors[new_phx_factor.fuel_name] = new_phx_factor
    return


def add_local_co2_conversion_factors(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Copy local Site->CO2e conversion factors from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the factor data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    for factor in _hb_room.properties.ph.ph_bldg_segment.co2e_factors: # type: ignore
        new_phx_factor = phx_site.PhxCO2Factor()
        new_phx_factor.fuel_name = factor.fuel_name
        new_phx_factor.value = factor.value
        new_phx_factor.unit = factor.unit
        _variant.site.energy_factors.co2_factors[new_phx_factor.fuel_name] = new_phx_factor
    return


def add_ventilation_systems_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Add new HVAC (Ventilation, etc) Systems to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    for hbph_space in _hb_room.properties.ph.spaces: # type: ignore
        # -- Get the Honeybee-PH Ventilation system from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        hbph_vent_sys: PhVentilationSystem = hbph_space.host.properties.energy.hvac.properties.ph.ventilation_system
        
        if not hbph_vent_sys:
            continue

        # -- Get or Build the PHX Ventilation Device
        # -- If the ventilator already exists, just use that one.
        phx_ventilator = _variant.mech_systems.get_mech_device_by_key(hbph_vent_sys.key)
        if not phx_ventilator:
            # -- otherwise, build a new PH-Ventilator from the HB-hvac
            phx_ventilator = create_hvac.build_phx_ventilation_sys(hbph_vent_sys)
            _variant.mech_systems.add_new_mech_device(hbph_vent_sys.key, phx_ventilator)
        
        # -- Re-set the HBPH Ventilator and sys id_num to keep everything aligned
        hbph_vent_sys.id_num = phx_ventilator.id_num
        
        if hbph_vent_sys.ventilation_unit:
            hbph_vent_sys.ventilation_unit.id_num = phx_ventilator.id_num

    return None


def add_heating_systems_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Add a new PHX-Heating SubSystem to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    for space in _hb_room.properties.ph.spaces:# type: ignore
        # -- Get the Honeybee-PH Heating Systems from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        heating_systems: Set[PhHeatingSystem] = space.host.properties.energy.hvac.properties.ph.heating_systems

        # -- Get or Build the PHX Heating systems
        # -- If the system already exists, just use that one.
        for hbph_sys in heating_systems:
            phx_heating_device = _variant.mech_systems.get_mech_device_by_key(hbph_sys.key)
            if not phx_heating_device:
                # -- otherwise, build a new PHX-Heating-Sys from the HB-hvac
                phx_heating_device = create_hvac.build_phx_heating_sys(hbph_sys)
                _variant.mech_systems.add_new_mech_device(hbph_sys.key, phx_heating_device)

            hbph_sys.id_num = phx_heating_device.id_num

    return None


def add_cooling_systems_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Add new PHX-Cooling SubSystem to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    for space in _hb_room.properties.ph.spaces:
        # -- Get the Honeybee-PH Cooling-Systems from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        cooling_systems: Set[PhCoolingSystem] = space.host.properties.energy.hvac.properties.ph.cooling_systems

        # -- Get or Build the PHX-Cooling systems
        # -- If the system already exists, just use that one.
        for hbph_sys in cooling_systems:
            phx_cooling_device = _variant.mech_systems.get_mech_device_by_key(hbph_sys.key)
            if not phx_cooling_device:
                # -- otherwise, build a new PHX-Cooling-Sys from the HB-hvac
                phx_cooling_device = create_hvac.build_phx_cooling_sys(hbph_sys)
                _variant.mech_systems.add_new_mech_device(hbph_sys.key, phx_cooling_device)

            hbph_sys.id_num = phx_cooling_device.id_num

    return None


def add_dhw_storage_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Add new Service Hot Water Equipment to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX Variant to add the PHX DHW System to.
        * _hb_room (room.Room): The Honeybee room to get the DHW System data from.

    Returns:
    --------
        * None
    """

    for space in _hb_room.properties.ph.spaces:# type: ignore
        # -- Guard Clause
        if not space.host.properties.energy.shw or not space.host.properties.energy.service_hot_water:
            continue

        # -- Build the HW-Tank
        for hw_tank in space.host.properties.energy.shw.properties.ph.tanks:
            if not hw_tank:
                continue

            if _variant.mech_systems.device_in_collection(hw_tank.key):
                continue

            # -- Build a new PHS-HW-Tank from the HB-hvac
            phx_dhw_tank = create_shw.build_phx_hw_storage(hw_tank)
            _variant.mech_systems.add_new_mech_device(hw_tank.key, phx_dhw_tank)

    return None


def add_dhw_heaters_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """
    
    Arguments:
    ----------
        *_variant (project.PhxVariant): The PHX Variant to add the PHX DHW Heaters to.
        _hb_room (room.Room): The Honeybee room to get the DHW Heater data from.
    
    Returns:
    --------
        * None
    """

    for space in _hb_room.properties.ph.spaces: # type: ignore

        if not space.host.properties.energy.shw:
            continue

        for heater in space.host.properties.energy.shw.properties.ph.heaters:
            if _variant.mech_systems.device_in_collection(heater.identifier):
                continue

            # -- Build a new PHX-HW-Heater from the Honeybee-PH HW-Heater
            phx_hw_heater = create_shw.build_phx_hw_heater(heater)
            _variant.mech_systems.add_new_mech_device(heater.identifier, phx_hw_heater)


def add_dhw_piping_from_hb_rooms(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    for space in _hb_room.properties.ph.spaces:# type: ignore

        if not space.host.properties.energy.shw:
            continue

        for branch_piping_element in space.host.properties.energy.shw.properties.ph.branch_piping:
            _variant.mech_systems.add_branch_piping(
                create_shw.build_phx_piping(branch_piping_element)
            )
        
        _variant.mech_systems._distribution_num_hw_tap_points = space.host.properties.energy.shw.properties.ph.number_tap_points

        for recirc_piping_element in space.host.properties.energy.shw.properties.ph.recirc_piping:
            _variant.mech_systems.add_recirc_piping(
                create_shw.build_phx_piping(recirc_piping_element)
            )

    return None


def add_elec_equip_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Creates new PHX-Elec-Equipment (Appliances) and adds them to each of the Variant.building.zones

    Arguments:
    ----------
        * _variant (project.Variant): The Variant to add the new elec-equipment to.
        # _hb_room (room.Room): The Honeybee Room to get the elec-equipment from.

    Returns:
    --------
        * None
    """

    ee_properties_ph: equipment.ElectricEquipmentPhProperties = _hb_room.properties.energy.electric_equipment.properties.ph# type: ignore
    for equip_key, device in ee_properties_ph.equipment_collection.items():
        phx_elec_device = create_elec_equip.build_phx_elec_device(device)
        for zone in _variant.building.zones:
            zone.elec_equipment_collection.add_new_device(
                equip_key, phx_elec_device)

    return


def from_hb_room(_hb_room: room.Room,
                 _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
                 _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
                 group_components: bool = False) -> project.PhxVariant:
    """Create a new PHX-Variant based on a single PH/Honeybee Room.

    Arguments:
    ----------
        * _hb_room (honeybee.room.Room): The honeybee room to base the Variant on.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.
        * group_components (bool): default=False. Set to true to have the converter
            group the components by assembly-type.

    Returns:
    --------
        * A new Variant object.
    """

    new_variant = project.PhxVariant()

    # -- Keep all the ID numbers aligned
    new_variant.id_num = project.PhxVariant._count
    _hb_room.properties.ph.id_num = new_variant.id_num# type: ignore
    new_variant.name = _hb_room.display_name

    # -- Build the Variant Elements (Dev. note: order matters!)
    add_ventilation_systems_from_hb_rooms(new_variant, _hb_room)
    add_heating_systems_from_hb_rooms(new_variant, _hb_room)
    add_cooling_systems_from_hb_rooms(new_variant, _hb_room)
    add_dhw_heaters_from_hb_rooms(new_variant, _hb_room)
    add_dhw_piping_from_hb_rooms(new_variant, _hb_room)
    add_dhw_storage_from_hb_rooms(new_variant, _hb_room)
    add_building_from_hb_room(new_variant, _hb_room, _assembly_dict,
                              _window_type_dict, group_components)
    add_phius_certification_from_hb_room(new_variant, _hb_room)
    add_phi_certification_from_hb_room(new_variant, _hb_room)
    add_PhxPhBuildingData_from_hb_room(new_variant, _hb_room)
    add_climate_from_hb_room(new_variant, _hb_room)
    add_local_pe_conversion_factors(new_variant, _hb_room)
    add_local_co2_conversion_factors(new_variant, _hb_room)
    add_elec_equip_from_hb_room(new_variant, _hb_room)

    return new_variant



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Mechanical Heating Devices"""

from __future__ import annotations
from typing import Optional, Union
from dataclasses import dataclass, field

from PHX.model.enums.hvac import DeviceType, HeatPumpType, FuelType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxHeatingDevice(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()


# -----------------------------------------------------------------------------
# Electric

@dataclass
class PhxHeaterElectricParams(_base.PhxMechanicalDeviceParams):
    pass


@dataclass
class PhxHeaterElectric(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.ELECTRIC)
    device_type: DeviceType = field(init=False, default=DeviceType.ELECTRIC)
    params: PhxHeaterElectricParams = field(
        default_factory=PhxHeaterElectricParams)


# -----------------------------------------------------------------------------
# Boilers


@dataclass
class PhxHeaterBoilerFossilParams(_base.PhxMechanicalDeviceParams):
    _fuel: FuelType = FuelType.GAS
    condensing: bool = True
    in_conditioned_space: bool = True
    effic_at_30_percent_load: float = 0.98
    effic_at_nominal_load: float = 0.94
    avg_rtrn_temp_at_30_percent_load: float = 30
    avg_temp_at_70C_55C: float = 41
    avg_temp_at_55C_45C: float = 35
    avg_temp_at_32C_28C: float = 24
    standby_loss_at_70C: Optional[float] = None
    rated_capacity: float = 10.0  # kW

    @property
    def fuel(self):
        return self._fuel

    @fuel.setter
    def fuel(self, _in):
        self._fuel = FuelType(_in)


@dataclass
class PhxHeaterBoilerWoodParams(_base.PhxMechanicalDeviceParams):
    _fuel: FuelType = FuelType.WOOD_LOG
    effic_in_basic_cycle: float = 0.6
    effic_in_const_operation: float = 0.7
    avg_frac_heat_output: float = 0.4
    temp_diff_on_off: float = 30.0
    rated_capacity: float = 15.0  # kW
    demand_basic_cycle: float = 1.0  # kWh
    power_stationary_run: float = 1.0  # W
    power_standard_run: Optional[float] = None
    no_transport_pellets: Optional[bool] = None
    only_control: Optional[bool] = None
    area_mech_room: Optional[float] = None

    @property
    def fuel(self):
        return self._fuel

    @fuel.setter
    def fuel(self, _in):
        self._fuel = FuelType(_in)


@dataclass
class PhxHeaterBoilerFossil(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.BOILER)
    device_type: DeviceType = field(init=False, default=DeviceType.BOILER)
    params: PhxHeaterBoilerFossilParams = field(
        default_factory=PhxHeaterBoilerFossilParams)


@dataclass
class PhxHeaterBoilerWood(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.BOILER)
    device_type: DeviceType = field(init=False, default=DeviceType.BOILER)
    params: PhxHeaterBoilerWoodParams = field(
        default_factory=PhxHeaterBoilerWoodParams)


PhxHeaterBoiler = Union[PhxHeaterBoilerFossil, PhxHeaterBoilerWood]


# -----------------------------------------------------------------------------
# District Heat


@dataclass
class PhxHeaterDistrictHeatParams(_base.PhxMechanicalDeviceParams):
    pass


@dataclass
class PhxHeaterDistrictHeat(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.DISTRICT_HEAT)
    device_type: DeviceType = field(init=False, default=DeviceType.DISTRICT_HEAT)
    params: PhxHeaterDistrictHeatParams = field(
        default_factory=PhxHeaterDistrictHeatParams)


# -----------------------------------------------------------------------------
# Heat Pumps


@dataclass
class PhxHeaterHeatPumpAnnualParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.ANNUAL)
    annual_COP: Optional[float] = None
    total_system_perf_ratio: Optional[float] = None


@dataclass
class PhxHeaterHeatPumpMonthlyParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.RATED_MONTHLY)
    COP_1: Optional[float] = None
    COP_2: Optional[float] = None
    ambient_temp_1: Optional[float] = None
    ambient_temp_2: Optional[float] = None

    @property
    def monthly_COPS(self):
        return None

    @monthly_COPS.setter
    def monthly_COPS(self, _in):
        if not _in:
            return

        self.COP_1 = _in[0]
        try:
            self.COP_2 = _in[1]
        except IndexError:
            self.COP_2 = _in[0]

    @property
    def monthly_temps(self):
        return None

    @monthly_temps.setter
    def monthly_temps(self, _in):
        if not _in:
            return

        self.ambient_temp_1 = _in[0]
        try:
            self.ambient_temp_2 = _in[1]
        except IndexError:
            self.ambient_temp_2 = _in[0]


@dataclass
class PhxHeaterHeatPumpHotWaterParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.HOT_WATER)
    annual_COP: Optional[float] = None
    annual_system_perf_ratio: Optional[float] = None
    annual_energy_factor: Optional[float] = None


@dataclass
class PhxHeaterHeatPumpCombinedParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.COMBINED)


@dataclass
class PhxHeaterHeatPumpAnnual(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeaterHeatPumpAnnualParams = field(
        default_factory=PhxHeaterHeatPumpAnnualParams)


@dataclass
class PhxHeaterHeatPumpMonthly(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeaterHeatPumpMonthlyParams = field(
        default_factory=PhxHeaterHeatPumpMonthlyParams)


@dataclass
class PhxHeaterHeatPumpHotWater(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeaterHeatPumpHotWaterParams = field(
        default_factory=PhxHeaterHeatPumpHotWaterParams)


@dataclass
class PhxHeaterHeatPumpCombined(PhxHeatingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeaterHeatPumpCombinedParams = field(
        default_factory=PhxHeaterHeatPumpCombinedParams)


PhxHeaterHeatPump = Union[PhxHeaterHeatPumpAnnual, PhxHeaterHeatPumpMonthly,
                          PhxHeaterHeatPumpHotWater, PhxHeaterHeatPumpCombined]

AnyPhxHeater = Union[
    PhxHeaterElectric,
    PhxHeaterBoilerFossil, 
    PhxHeaterBoilerWood,
    PhxHeaterDistrictHeat,
    PhxHeaterHeatPumpAnnual,
    PhxHeaterHeatPumpMonthly,
    PhxHeaterHeatPumpHotWater,
    PhxHeaterHeatPumpCombined,
]


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for importing / converting WUFI-XML into PHX Models"""

import pathlib


def read_WUFI_XML_from_file(_file_address: pathlib.Path) -> str:
    """Read in the WUFI-XML file and return it as a string.

    Arguments:
    ----------
        _file_address (pathlib.Path): A valid file path for the WUFI-XML file to read.

    Returns:
    --------
        str: The WUFI XML text, read in from the WUFI-XML file.
    """

    with open(_file_address) as xml_file:
        data = xml_file.read()

    return data



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Class for managing the XL-Connection and common read/write operations."""

from typing import Optional, List, Set, Callable, Any
from contextlib import contextmanager
import os

import xlwings as xw

from PHX.to_PHPP import xl_data

# -----------------------------------------------------------------------------

class SupportsString:
    """An object which can be stringified."""

    def __str__(self): ...

# -----------------------------------------------------------------------------


class ReadRowsError(Exception):
    def __init__(self, row_start, row_end):
        self.msg = f'Error: row_start should be less than row_end. Got {row_start}, {row_end}'
        super().__init__(self.msg)


class NoActiveExcelRunningError(Exception):
    def __init__(self):
        self.msg = '\n\tError: No active instance of Excel running? Please open a valid PHPP file and try again.'
        super().__init__(self.msg)


class ReadMultipleColumnsError(Exception):
    def __init__(self, _c1, _c2):
        self.msg = f'\n\tError: Cannot use "read_multiple_columns()" with _col_start={_c1}'\
            f'and _col_end={_c2}. Please use "read_single_column()" instead.'
        super().__init__(self.msg)


class WriteValueError(Exception):
    def __init__(self, _value, _range, _worksheet, _e):
        self.msg = "\n\n\tSomething went wrong trying to write the value: '{}' to the cell: '{}' on worksheet: '{}'. Please "\
                    "make sure that a valid PHPP file is open, and both the worksheet and workbook are unprotected.\n\n{}".format(
                         _value, _range, _worksheet, _e
                    )
        super().__init__(self.msg)


# -----------------------------------------------------------------------------


class XLConnection:

    def __init__(self, _output: Optional[Callable]=None):
        """Facade class for Excel Interop
            
        Arguments:
        ----------
            * _log: Optional[Callable]: The output functions to use. Input None for silent.
        """
    
        self._output = _output

    def output(self, _input: SupportsString):
        """Used to set the output method. Default is None (silent).
        
        Arguments:
        ----------
            * _input (SupportsString): The string to output.
        
        Returns:
        --------
            *
        """
        try:
            self._output(str(_input))
        except:
            pass

    @property
    def wb(self) -> xw.main.Book:
        try:
            return xw.books.active
        except:
            raise NoActiveExcelRunningError

    @contextmanager
    def in_silent_mode(self):
        """Context Manager which turns off screen-refresh and auto-calc in the
            Excel App in order to help speed up read/write. Turns back on screen-refresh
            and auto-calc in the Excel App when done or on any error.
        """
        try:
            self.wb.app.screen_updating = False
            self.wb.app.display_alerts = False
            self.wb.app.calculation = 'manual'
            yield
        finally:
            self.wb.app.screen_updating = True
            self.wb.app.display_alerts = True
            self.wb.app.calculation = 'automatic'
            self.wb.app.calculate()

    def connection_is_open(self) -> bool:
        if not self.wb:
            return False
        return True

    def unprotect_all_sheets(self) -> None:
        """Walk through all the sheets and unprotect them all."""
        for sheet in self.wb.sheets:
            if os.name != 'nt':
                sheet.api.unprotect()
            else:
                sheet.api.Unprotect()

    def get_row_num_of_value_in_column(self, sheet_name: str, row_start: int,
                                       row_end: int,  col: str, find: str) -> Optional[int]:
        """Returns the row number of the first instance of a specific value found within a column, or None if not found.

        Arguments:
        ----------
            * sheet_name: (str) The name of the sheet to be looking in
            * row_start: (int) The row number to begin looking from
            * row_end: (int) The row number to stop looking on
            * col: (str) The column to look in
            * find: (Optional[str]) The string to search for (or 'None' for blank cell)

        Returns:
        --------
            * Optional[int]: The row number where the first instance of the specified 
                value is found or None if no instances are found.
        """

        if row_start > row_end:
            raise ReadRowsError(row_start, row_end)

        row: int = row_start
        sh: xw.main.Sheet = self.get_sheet_by_name(sheet_name)

        while row <= row_end:
            if sh.range(f'{col}{row}').value != find:
                row += 1
            else:
                return row
        return None

    def get_worksheet_names(self) -> Set[str]:
        """Return a list of all the worksheet names in the workbook.
            
        Returns:
        --------
            * (List[str])
        """
        return {sh.name for sh in self.wb.sheets}

    def get_sheet_by_name(self, _sheet_name: str) -> xw.main.Sheet:
        """Returns an Excel Sheet with the specified name, or KeyError if not found.

        Arguments:
        ----------
            * _sheet_name: (str): The excel sheet name to locate.

        Returns:
        --------
            * (xw.main.Sheet): The excel sheet found.
        """
        return self.wb.sheets[_sheet_name]

    def get_single_column_data(self, _sheet_name: str, _col: str, _row_start: int = 1,
                               _row_end: int = 100, ) -> List[xl_data.xl_range_value]:
        """Return a list with the values read from a single column of the excel document.

        Arguments:
        ----------
            * _sheet_name: (str) The Excel Worksheet to read from.
            * _col: (str) The Column letter to read.
            * _row_start: (int) default=1
            * _row_end: (int) default=100
        Returns:
        --------
            (List[xl_range_value]): The data from Excel worksheet, as a list.
        """

        return self.get_sheet_by_name(_sheet_name).range(f'{_col}{_row_start}:{_col}{_row_end}').value

    def get_single_row_data(self, _sheet_name: str, _row_number: int) -> List[xl_data.xl_range_value]:
        """
        
        Arguments:
        ----------
            * _sheet_name (str): The name of the sheet to read
            * _row_number (int): The row to read
        
        Returns:
        --------
            * (List[xl_data.xl_range_value]) The data read from XL.
        """

        sht = self.get_sheet_by_name(_sheet_name)
        return sht.range((_row_number, 1),(_row_number, 500)).value

    def get_multiple_column_data(self, _sheet_name: str, _col_start: str, _col_end: str,
                                 _row_start: int = 1, _row_end: int = 100, ) -> List[List[xl_data.xl_range_value]]:
        """Return a list with the values read from a specified block of the xl document.

        Arguments:
        ----------
            * _sheet_name: (str) The Worksheet to read from.
            * _col_start: (str) The Column letter to read from.
            * _col_end: (str) The Column letter to read to.
            * _row_start: (int) default=1
            * _row_end: (int) default=100

        Returns:
        --------
            (List[List[xl_range_value]]): The data from Excel worksheet, as a list of lists.
        """

        if _col_start == _col_end:
            raise ReadMultipleColumnsError(_col_start, _col_end)

        # -- Use XW instead of ord() since ord('KL') and similar will fail
        rng = xw.Range(f'{_col_end}1:{_col_start}1')
        _ndim = len(rng.columns)

        return self.get_sheet_by_name(_sheet_name).range(
            f'{_col_start}{_row_start}:{_col_end}{_row_end}').options(ndim=_ndim).value

    def get_data(self, _sheet_name: str, _range: str) -> xl_data.xl_writable:
        """Return a value or values from Excel

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to read from.
            * _range: (str) The cell range to write to (ie: "A1") or a set of ranges (ie: "A1:B4")

        Returns:
        ---------
            * (xl_writable): The resultant value/values returned from excel.
        """
        self.output(f'Reading: {_sheet_name}:{_range}')
        return self.get_sheet_by_name(_sheet_name).range(_range).value

    def write_xl_item(self, _xl_item: xl_data.XlItem) -> None:
        """Writes a single XLItem to the worksheet

        Arguments:
        ---------
            * _xl_item: (XLItem) The XLItem with a sheet_name, range and value to write.
         """
        try:
            self.output(f'Writing {_xl_item.write_value} to {_xl_item.sheet_name}:{_xl_item.xl_range}')
            self.get_sheet_by_name(_xl_item.sheet_name).range(
                _xl_item.xl_range).value = _xl_item.write_value
        except Exception as e:
            raise WriteValueError(_xl_item.write_value, _xl_item.xl_range, _xl_item.sheet_name, e)

    def clear_data(self, _sheet_name: str, _range: str) -> None:
        """Sets the specified excel sheet's range to 'None'

        Arguments:
        ---------
            * _sheet_name: (str) The name of the sheet to write to
            * _range: (str) The cell range to write to (ie: "A1") or a set of ranges (ie: "A1:B4")
         """
        self.get_sheet_by_name(_sheet_name).range(_range).value = None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

""""""
from typing import Optional, Callable, List, Any

from PHX.to_WUFI_XML import xml_schemas
from PHX.to_WUFI_XML.xml_writables import xml_writable


class NoXMLSchemaFoundError(Exception):
    def __init__(self, _schema_module, _phx_object, _schema_nm):
        self.message = (
            f'\n  Error: Cannot find an XML write schema for the object: "{_phx_object}"'
            f'\n  of type "{type(_phx_object)}" using the schema name of: "{_schema_nm}"'
            f'\n  in file: "{_schema_module.__file__}". Please check the schemas.'
        )
        super(NoXMLSchemaFoundError, self).__init__(self.message)


def get_PHX_object_conversion_schema(_phx_object, _schema_name: Optional[str] = None) -> Callable[[Any], List[xml_writable]]:
    """Returns the appropriate XML write schema function for the PHX-object.

    Arguments:
    ----------
        * _phx_object (Any): The PHX-Object to find the WUFI-XML write schema for.
        * _schema_name (Optional[str]): Optional user-defined name of the XML schema to use.
            If None is supplied, will use the object name preceded by an underscore. Ie:
            "Room" will search for "_Room".
    Returns:
    --------
        * (Callable[[Any], List[xml_writable]): The conversion function
            which can be used to convert the PHX-Object into WUFI-XML Format.

    Raises:
    -------
        * (NoXMLSchemaFoundError): If no valid XML conversion schema is found for the 
            designated object.
    """

    # -- Schema Name
    if _schema_name is None:
        _schema_name = "_{}".format(_phx_object.__class__.__name__)

    # -- Schema Function
    schema_function = getattr(xml_schemas, _schema_name, None)
    if not schema_function:
        raise NoXMLSchemaFoundError(
            xml_schemas, _phx_object, _schema_name)

    return schema_function


def convert_HB_object_to_xml_writables_list(_phx_object, _schema_nm: Optional[str] = None) -> List[xml_writable]:
    """Returns a list of the PHX-Object's Properties in WUFI-XML format.

        * _phx_object (Any): The PHX-Object to convert into XML text.
        * _schema_nm (Optional[str]): The name of the conversion schema to use for the object.

    Returns:
    --------
        * (List[xml_writable]):

        ie: [
            XML_Node('IdentNr', 2),
            XML_Node('Name', 'My-Object'),
            XML_List( ... ),
            ...
            ]
    """

    # -- Get the right XML conversion function for the object
    conversion_schema = get_PHX_object_conversion_schema(_phx_object, _schema_nm)

    # # -- Convert the object to an XML Node List
    return conversion_schema(_phx_object)



"""PHX Mechanical Equipment (water, heating, cooling, ventilation)"""

from PHX.model.enums.hvac import *
from PHX.model.hvac._base import *
from PHX.model.hvac.collection import *
from PHX.model.hvac.cooling import *
from PHX.model.hvac.heating import *
from PHX.model.hvac.ventilation import *
from PHX.model.hvac.water import *
from PHX.model.hvac.piping import *



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to cleanup / optimize Honeybee-Rooms before outputting to WUFI"""

from typing import List, Union, Tuple
from functools import partial

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_ep_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:
    from honeybee import room, face
    from honeybee.boundarycondition import Outdoors, Ground
    from honeybee_energy.boundarycondition import Adiabatic
    from honeybee_energy import shw
    from honeybee_energy.load import people, equipment, infiltration
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
    from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

from honeybee_energy_ph.properties.hot_water.hw_system import SHWSystemPhProperties
from PHX.model import project


HB_BC = Union[Outdoors, Ground, Adiabatic]


def _get_room_exposed_faces(_hb_room: room.Room, _bc_types: Tuple = (Outdoors, Ground, Adiabatic)) -> List[face.Face3D]:
    """Returns a list of the Honeybee Faces within the set of HB-Boundary Conditions specified.

    Arguments:
    ----------
        * _hb_room (room.Room): The Honeybee Room to get the faces of.
        * _type (tuple[HB_BC]): A tuple of the allowable HB-Boundary Conditions.
            default = (Outdoors, Ground, Adiabatic)

    Returns:
    --------
        * list[face.Face]: The list of Exposed Honeybee Faces.
    """
    
    exposed_faces = []
    for original_face in _hb_room.faces:
        if not isinstance(original_face.boundary_condition, _bc_types):
            continue

        new_face = original_face.duplicate()
        new_face._properties._duplicate_extension_attr(original_face._properties)
        
        # -- Note, this is required if the user has set custom .energy constructions
        # -- or other custom face-specific attributes
        # -- Duplicate any extensions like .ph, .energy or .radiance
        for extension_name in original_face.properties._extension_attributes:
            # -- use try except to avoid the bug introduced in HB v1.5
            # TODO: fix Honeybee and then remove this try...except...
            try:
                original_extension = getattr(original_face._properties, f'_{extension_name}', None)
                new_extension = original_extension.duplicate()
                setattr(new_face._properties, f'_{extension_name}', new_extension)
            except:
                pass

        exposed_faces.append(new_face)

    return exposed_faces


def _get_room_exposed_face_area(_hb_room: room.Room, _bc_types: Tuple) -> float:
    """Returns the total area of the 'exposed' faces in a HB-Room."""
    return sum(f.area for f in _get_room_exposed_faces(_hb_room, _bc_types))


def merge_occupancies(_hb_rooms: List[room.Room]) -> people.People:
    """Returns a new HB-People-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-People object from.

    Return:
    ------- 
        * (people.People): A new Honeybee People object with values merged from the HB-Rooms.
    """

    # Tally up all the values from all the rooms
    total_hb_people = 0.0
    total_ph_bedrooms = 0.0
    total_ph_people = 0.0
    for room in _hb_rooms:
        hb_ppl_obj = room.properties.energy.people  # alias
        total_ph_bedrooms += int(hb_ppl_obj.properties.ph.number_bedrooms)
        total_ph_people += int(hb_ppl_obj.properties.ph.number_people)
        total_hb_people += hb_ppl_obj.people_per_area * room.floor_area

    # Build up the new object's attributes
    total_floor_area = sum(rm.floor_area for rm in _hb_rooms)
    new_hb_ppl = _hb_rooms[0].properties.energy.people.duplicate() # type: ignore
    new_hb_ppl.people_per_area = total_hb_people / total_floor_area
    new_hb_ppl.properties.ph.number_bedrooms = total_ph_bedrooms
    new_hb_ppl.properties.ph.number_people = total_ph_people

    return new_hb_ppl


def merge_infiltrations(_hb_rooms: List[room.Room]) -> infiltration.Infiltration:
    """Returns a new HB-Infiltration-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-Infiltration object from.

    Return:
    ------- 
        * (infiltration.Infiltration): A new Honeybee Infiltration object 
            with values merged from the HB-Rooms.
    """
    # For Phius, infiltration-exposed surfaces include all, including 'Ground'
    # unlike for Honeybee where only 'Outdoors' count as 'exposed'
    get_rm_infil_exposed_face_area = partial(
        _get_room_exposed_face_area, _bc_types=(Outdoors, Ground))

    # Calculate the total airflow per room, total exposed area per room
    total_m3_s = 0.0
    total_exposed_area = 0.0
    for room in _hb_rooms:
        room_infil_exposed_area = get_rm_infil_exposed_face_area(room)
        room_infil_m3_s = room_infil_exposed_area * \
            room.properties.energy.infiltration.flow_per_exterior_area

        total_exposed_area += room_infil_exposed_area
        total_m3_s += room_infil_m3_s

    # -- Set the new Infiltration Object's attr to the weighted average
    new_infil = _hb_rooms[0].properties.energy.infiltration.duplicate()
    try:
        new_infil.flow_per_exterior_area = total_m3_s / total_exposed_area
    except ZeroDivisionError:
        new_infil.flow_per_exterior_area = 0.0

    return new_infil


def merge_shw(_hb_rooms: List[room.Room]) -> shw.SHWSystem:
    """Merge together several HB-Room's Honeybee-Energy SHW System objects (if they exist).
    
    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): The list of Honeybee Rooms to get the 
            SHW System System from.
    
    Returns:
    --------
        * (shw.SHWSystem): A single new Honeybee-Energy SHW System Object.   
    """

    # -- Find the first HBE SHWSystem System in the list of HB-Rooms, and use that as the 'base'
    for room in _hb_rooms:
        if room.properties.energy.shw: #type: ignore
            new_shw: shw.SHWSystem = room.properties.energy.shw.duplicate() # type: ignore
            break
    else:
        # -- If no SHWSystem found anywhere, return a new default HBE SHWSystem System
        return shw.SHWSystem(
            identifier=clean_ep_string('_default_shw_system_'),
            equipment_type='Electric_WaterHeater',
        )

    # -- Merge the SHWSystemPhProperties and then apply it to the new HBE SHWSystem System
    shw_props: List[SHWSystemPhProperties] = [hb_room.properties.energy.shw.properties.ph # type: ignore
                                                for hb_room in _hb_rooms 
                                                if hb_room.properties.energy.shw] # type: ignore
    new_prop = sum(shw_props) # type: ignore
    new_shw.properties._ph = new_prop # type: ignore

    return new_shw


def merge_elec_equip(_hb_rooms: List[room.Room]) -> equipment.ElectricEquipment:
    """Returns a new HB-ElectricEquipment-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-ElectricEquipment object from.

    Return:
    ------- 
        * (equipment.ElectricEquipment): A new Honeybee ElectricEquipment object 
            with values merged from the HB-Rooms.
    """

    # -- Collect all the unique PH-Equipment in all the rooms.
    # -- Increase the quantity for each duplicate piece of equipment found.
    ph_equipment = {}
    for room in _hb_rooms:
        for equip_key, equip in room.properties.energy.electric_equipment.properties.ph.equipment_collection.items():
            try:
                ph_equipment[equip_key].quantity += 1
            except KeyError:
                ph_equipment[equip_key] = equip

    # -- Calculate the total Watts of elec-equipment, total floor-area
    total_floor_area = sum(rm.floor_area for rm in _hb_rooms)
    total_watts = sum(
        (rm.floor_area * rm.properties.energy.electric_equipment.watts_per_area) for rm in _hb_rooms)

    # -- Build a new HB-Elec-Equip with all the PH-Equipment in it.
    new_hb_equip = _hb_rooms[0].properties.energy.electric_equipment.duplicate()
    new_hb_equip.watts_per_area = total_watts / total_floor_area
    new_hb_equip.properties.ph.equipment_collection.remove_all_equipment()
    for ph_item in ph_equipment.values():
        new_hb_equip.properties.ph.equipment_collection.add_equipment(ph_item)

    return new_hb_equip


def merge_rooms(_hb_rooms: List[room.Room]) -> room.Room:
    """Merge together a group of Honeybee Rooms into a new single HB Room.

    This will
    ignore any 'interior' Honeybee-Faces with a 'Surface' boundary condition and will only
    keep the 'exposed' Honeybee Faces with boundary_conditions of 'Outdoors', 'Ground'
    and 'Adiabatic' to build the new Honeybee Room from.

    Arguments:
    ----------
        * _hb_rooms (list[room.Room]):

    Returns:
    --------
        * room.Room: The new Honeybee Room.
    """
    reference_room = _hb_rooms[0]
    
    # -------------------------------------------------------------------------
    # -- Get only the 'exposed' faces to build a new HB-Room with
    exposed_faces = []
    for hb_room in _hb_rooms:
        exposed_faces += _get_room_exposed_faces(hb_room)

    new_room = room.Room(
        identifier=reference_room.properties.ph.ph_bldg_segment.display_name,
        faces=exposed_faces,
    )
    
    # -------------------------------------------------------------------------
    # -- Set the new Merged-Room's properties.ph and
    # -- properties.energy to match the 'reference' room to start with, but leave
    # -- off the PH-Spaces so they don't get duplicated
    dup_ph_prop = reference_room._properties.ph.duplicate(
        new_room._properties.ph, include_spaces=False)
    setattr(new_room._properties, '_ph', dup_ph_prop)

    dup_energy_prop = reference_room._properties.energy.duplicate(
        new_room._properties.energy)
    setattr(new_room._properties, '_energy', dup_energy_prop)

    # -------------------------------------------------------------------------
    # -- Then, collect all the spaces from the input rooms and add to the NEW room
    # -- NOTE: this has to be done AFTER the duplicate()
    # -- call, otherwise not all the spaces will transfer over properly.
    for hb_room in _hb_rooms:
        for existing_space in hb_room.properties.ph.spaces:
            # -- Preserve the original HB-Room's energy and ph properties over
            # -- on the space. We need to do this cus' the HB-Room is being removed
            # -- and we want to preserve HVAC and program info for the spaces.
            existing_space.properties._energy = hb_room.properties._energy.duplicate(
                new_host=existing_space)
            # TODO: Verify that this can be removed without causing any issues?
            # Note: it it also wrong - RoomPhProperties are being applied over SpacePhProperties.
            # existing_space.properties._ph = hb_room.properties._ph.duplicate(
            #     new_host=existing_space)
            
            new_room.properties.ph.add_new_space(existing_space)

    # -------------------------------------------------------------------------
    # -- Merge the hb_rooms' load values
    new_room.properties.energy.infiltration = merge_infiltrations(_hb_rooms)
    new_room.properties.energy.people = merge_occupancies(_hb_rooms)
    new_room.properties.energy.electric_equipment = merge_elec_equip(_hb_rooms)
    new_room.properties.energy.shw = merge_shw(_hb_rooms)

    # -------------------------------------------------------------------------
    # -- TODO: Can I merge together the surfaces as well?
    # -- For larger models, I think this will be important.... hmm....
    # -- Organize the surfaces by -> assembly / exposure / orientation (normal)

    return new_room


def weld_vertices(_variant: project.PhxVariant) -> project.PhxVariant:
    """
    Used to try and weld/unify the vertices of a variant.

    This is helpful to reduce the complexity / number of variants in a large
    model so that WUFI can actually open it.

    Arguments:
    ----------
        * _variant (project.Variant): The Variant object to weld the vertices for.

    Returns:
    --------
        * (project.Variant): The variant, with its vertix objects welded.

    """

    unique_vertix_dict = {}
    for component in _variant.building.all_components:
        for polygon in component.polygons:
            for i, vert in enumerate(polygon.vertices):
                try:
                    vert = polygon.vertices[i] = unique_vertix_dict[vert.unique_key]
                except KeyError:
                    unique_vertix_dict[vert.unique_key] = vert

    return _variant



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Data-entry constructor for the U-Values Worksheet."""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model import constructions

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class ConstructorBlock:
    """A single U-Value/Constructor entry block."""

    __slots__ = ("shape", "phx_construction")
    shape: shape_model.UValues
    phx_construction: constructions.PhxConstructionOpaque

    def _create_range(self, _field_name: str, _row_offset: int, _start_row: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.constructor.inputs, _field_name).column
        return f'{col}{_start_row + _row_offset}'
    
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.constructor.inputs, _field_name).unit

    def is_mass_material(self, _layer: constructions.PhxLayer) -> bool:
        """Return True if the Layer is an SD 'Mass' Layer."""
        
        for material in _layer.materials:
            if material.display_name == "MAT_Mass" and material.conductivity == 100:
                return True
        
        return False

    def create_xl_items(self, _sheet_name: str, _start_row: int) -> List[xl_data.XlItem]:
        """Convert the PHX-Construction into a list of XLItems for writing to the PHPP."""
                
        create_range = partial(self._create_range, _start_row=_start_row)
        XLItemUValues = partial(xl_data.XlItem, _sheet_name)

        # -- Build the basic assembly attributes
        xl_items_list: List[xl_data.XlItem] = [
            XLItemUValues(create_range('display_name', 2), self.phx_construction.display_name),
            XLItemUValues(create_range('r_si', 4), 0.0, "M2K/W", self._get_target_unit('r_si')),
            XLItemUValues(create_range('r_se', 5), 0.0, "M2K/W", self._get_target_unit('r_se')),
        ]

        # -- Build all the layers of the assembly
        for i, layer in enumerate(self.phx_construction.layers, start=8):
            if self.is_mass_material(layer):
                continue
            
            xl_items_list.append(XLItemUValues(create_range(
                        'thickness', i),
                        layer.thickness_mm,
                        "MM",
                        self._get_target_unit('thickness')
                    )
                )
            
            # -- add in each of the PhxMaterials found in the PhxLayer.materials collection
            for j, material in enumerate(layer.materials, start=1):
                layer_items: List[xl_data.XlItem] = [
                    XLItemUValues(create_range(f'sec_{j}_description', i),
                        material.display_name),
                    XLItemUValues(create_range(f'sec_{j}_conductivity', i),
                        material.conductivity,
                        "W/MK",
                        self._get_target_unit(f'sec_{j}_conductivity')
                    ),
                ]
                # -- Only add the percentage info for sections 2 and 3...
                if j > 1:
                    layer_items.append(
                        XLItemUValues(create_range(
                                f'sec_{j}_percentage', 
                                getattr(self.shape.constructor.inputs, f'sec_{j}_percentage').row
                                ),
                                material.percentage_of_assembly
                            )
                        )
                xl_items_list.extend(layer_items)

        return xl_items_list




# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Shading/Shading-Entry row"""

from dataclasses import dataclass
from typing import List, Tuple
from functools import partial
from PHX.model import components

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class ShadingRow:
    """Model class for a single Window's Shading entry row."""

    __slots__ = ('shape', 'shading_dims', 'winter_shading_factor', 'summer_shading_factor')
    shape: shape_model.Shading
    shading_dims: components.PhxApertureShadingDimensions
    winter_shading_factor: float
    summer_shading_factor: float

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.shading_rows.inputs, _field_name).column
        return f'{col}{_row_num}'

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.shading_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemShading = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemShading(create_range('h_hori'), self.shading_dims.h_hori, "M", self._get_target_unit("h_hori")),
            XLItemShading(create_range('d_hori'), self.shading_dims.d_hori, "M", self._get_target_unit("d_hori")),
            XLItemShading(create_range('o_reveal'), self.shading_dims.o_reveal, "M", self._get_target_unit("o_reveal")),
            XLItemShading(create_range('d_reveal'), self.shading_dims.d_reveal, "M", self._get_target_unit("d_reveal")),
            XLItemShading(create_range('o_over'), self.shading_dims.o_over, "M", self._get_target_unit("o_over")),
            XLItemShading(create_range('d_over'), self.shading_dims.d_over, "M", self._get_target_unit("d_over")),
            XLItemShading(create_range('r_other_winter'), self.winter_shading_factor),
            XLItemShading(create_range('r_other_summer'), self.summer_shading_factor),
        ]

        return items



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create new Shade PhxComponents from HB-Model Orphaned-Shade Objects."""

from honeybee import model, shade

from PHX.model import components, project
from PHX.model.enums.building import ComponentExposureExterior, ComponentFaceOpacity, ComponentColor
from PHX.from_HBJSON import create_geometry


def create_new_component_from_orphaned_shade(_shade: shade.Shade) -> components.PhxComponentOpaque:
    """Returns a new PHX-Component for based on the Honeybee-Shade.

    Arguments:
    ----------
        * _shade (shade.Shade): The Honeybee-Shade to base the new component on.

    Returns:
    --------
        * (components.PhxComponentOpaque): A new PHX-Component for the HB-Shade.
    """

    new_compo = components.PhxComponentOpaque()

    new_compo.display_name = _shade.display_name
    new_compo.face_opacity = ComponentFaceOpacity.OPAQUE
    new_compo.exposure_exterior = ComponentExposureExterior.EXTERIOR
    new_compo.exposure_interior = -1
    new_compo.color_interior = ComponentColor.EXT_WALL_INNER
    new_compo.color_exterior = ComponentColor.EXT_WALL_INNER

    # -- Polygons
    phx_polygon = create_geometry.create_PhxPolygon_from_hb_Face(_shade)
    new_compo.add_polygons(phx_polygon)

    return new_compo


def add_hb_model_shades_to_variant(_var: project.PhxVariant, _hb_model: model.Model) -> None:
    """"Create shading PhxComponents from an HB-model's orphaned shades and add to the PhxVariant.

    Arguments:
    ----------
        * _var (project.Variant): The PhxVariant to add the Shading Objects to.
        * _hb_model (model.Model): The Honeybee-Model to get the orphaned shades from.

    Returns:
    --------
        * None
    """

    for hb_shade_face in _hb_model.orphaned_shades:
        _var.building.add_components(
            create_new_component_from_orphaned_shade(hb_shade_face)
        )

    return None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to create Project elements from the Honeybee-Model"""

from typing import Union

from honeybee import model
from honeybee_energy.material.opaque import EnergyMaterial, EnergyMaterialNoMass
from honeybee_energy.construction import window

from honeybee_energy_ph.properties.construction.window import WindowConstructionPhProperties
from honeybee_ph_utils import iso_10077_1

from PHX.model import constructions, project


def _conductivity_from_r_value(_r_value: float, _thickness: float) -> float:
    """Returns a material conductivity value (W/mk), given a known r-value (M2K/W) and thickness (M).

    Arguments:
    ----------
        * _r_value (float): The total R-Value of the layer.
        * _thickness (float): The total thickness of the layer.

    Returns:
    --------
        * float: The Conductivity value of the layer.
    """
    conductivity = _thickness / _r_value
    return conductivity


def build_phx_material_from_hb_EnergyMaterial(_hb_material: EnergyMaterial) -> constructions.PhxMaterial:
    """Returns a new PhxMaterial, with attributes based on a Honeybee-Energy EnergyMaterial.

    Arguments:
    ----------
        * _hb_material (EnergyMaterialNoMass): The Honeybee-Energy EnergyMaterialNoMass
        * _thickness (float): The thickness of the material layer. 

    Returns:
    --------
        * (constructions.PhxMaterial): The new PhxMaterial with attributes based on the
            Honeybee-Energy EnergyMaterial.
    """
    new_mat = constructions.PhxMaterial()
    new_mat.display_name = _hb_material.display_name
    new_mat.conductivity = _hb_material.conductivity
    new_mat.density = _hb_material.density
    new_mat.heat_capacity = _hb_material.specific_heat
    new_mat.percentage_of_assembly = _hb_material.properties.ph.percentage_of_assembly # type: ignore
    
    # -- Defaults
    new_mat.porosity = 0.95
    new_mat.water_vapor_resistance = 1.0
    new_mat.reference_water = 0.0

    return new_mat


def build_phx_material_from_hb_EnergyMaterialNoMass(_hb_material: EnergyMaterialNoMass, _thickness: float) -> constructions.PhxMaterial:
    """Returns a new PhxMaterial, with attributes based on a Honeybee-Energy EnergyMaterialNoMass.
    Note that a thickness is required in order to determine the material conductivity.

    Arguments:
    ----------
        * _hb_material (EnergyMaterialNoMass): The Honeybee-Energy EnergyMaterialNoMass
        * _thickness (float): The thickness of the material layer. 

    Returns:
    --------
        * (constructions.PhxMaterial): The new PhxMaterial with attributes based on the
            Honeybee-Energy EnergyMaterialNoMass.
    """
    new_mat = constructions.PhxMaterial()
    new_mat.display_name = _hb_material.display_name
    new_mat.conductivity = _conductivity_from_r_value(_hb_material.r_value, _thickness)
    new_mat.density = _hb_material.mass_area_density
    new_mat.heat_capacity = _hb_material.area_heat_capacity

    # -- Defaults
    new_mat.water_vapor_resistance = 1.0
    new_mat.porosity = 0.95
    new_mat.reference_water = 0.0

    return new_mat


def build_layer_from_hb_material(_hb_material: Union[EnergyMaterial, EnergyMaterialNoMass]) -> constructions.PhxLayer:
    """Returns a new PHX-Layer with attributes based on a Honeybee-Material.

    Arguments:
    ----------
        *_hb_material (EnergyMaterial | EnergyMaterialNoMass): The Honeybee Material.

    Returns:
    --------
        * constructions.Layer: The new PHX-Layer object.
    """
    new_layer = constructions.PhxLayer()

    if isinstance(_hb_material, EnergyMaterial):
        new_layer.thickness_m = _hb_material.thickness

        if _hb_material.properties.ph.base_materials: # type: ignore
            ## -- Layer is a heterogeneous material layer (ie: stud + insulation)
            for hb_sub_material in _hb_material.properties.ph.base_materials: # type: ignore
                new_layer.add_material( build_phx_material_from_hb_EnergyMaterial(hb_sub_material) )
        else:
            # -- Layer is homogeneous, use the base HB Material
            new_layer.add_material( build_phx_material_from_hb_EnergyMaterial(_hb_material) )

    elif isinstance(_hb_material, EnergyMaterialNoMass):
        # -- NoMass layers cannot be heterogeneous (yet....)
        new_layer.thickness_m = 0.1  # 0.1m = 4". Use as default since No-Mass has no thickness
        new_layer.add_material( build_phx_material_from_hb_EnergyMaterialNoMass(_hb_material, new_layer.thickness_m) )

    else:
        raise TypeError(
            f"Error: Unsupported Material type: '{type(_hb_material)}'.")

    return new_layer


def build_opaque_assemblies_from_HB_model(_project: project.PhxProject, _hb_model: model.Model) -> None:
    """Build Opaque Constructions from Honeybee Faces and add to the PHX-Project.

    Will also align the id_nums of the face's Construction with the Assembly in the Project dict.

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to use as the source.

    Returns:
    --------
        * None
    """

    for room in _hb_model.rooms:
        for face in room.faces:
            hb_const = face.properties.energy.construction

            if not hb_const.identifier in _project.assembly_types:
                # -- Create a new Assembly with Layers from the Honeybee-Construction
                new_assembly = constructions.PhxConstructionOpaque()
                new_assembly.id_num = constructions.PhxConstructionOpaque._count
                new_assembly.display_name = hb_const.display_name
                new_assembly.layers = [build_layer_from_hb_material(layer)
                                       for layer in hb_const.materials]

                # -- Add the assembly to the Project
                _project.assembly_types[hb_const.identifier] = new_assembly

            hb_const.properties.ph.id_num = _project.assembly_types[hb_const.identifier].id_num

    return None


def build_phx_window_type_from_hb_win_construction(_hb_win_const: window.WindowConstruction) -> constructions.PhxConstructionWindow:
    """Create a new PHX-WindowType based on a HB-Window-Construction.

    If any detailed PH-Params exist for the frame or glass on the HB-Window-Construction's 
    .properties.ph.* then those will be used. Otherwise, the basic HB-Window-Construction
    attributes will be used.

    Arguments:
    ----------
        * _hb_win_const (window.WindowConstruction): The Honeybee Window Construction to
            base the new PHX-WindowType on.

    Returns:
    --------
        * (constructions.WindowType): The new PHX-WindowType.
    """

    phx_window_type = constructions.PhxConstructionWindow()
    phx_window_type.id_num = constructions.PhxConstructionWindow._count
    phx_window_type.display_name = _hb_win_const.display_name

    ph_params: WindowConstructionPhProperties = _hb_win_const.properties.ph

    # -- Glass ----------------------------------------------------------------
    if ph_params.ph_glazing:
        # -- Use Detailed PH-Params
        phx_window_type.glazing_type_display_name = ph_params.ph_glazing.display_name
        phx_window_type.u_value_glass = ph_params.ph_glazing.u_factor
        phx_window_type.glass_g_value = ph_params.ph_glazing.g_value
    else:
        # -- Use the basic Honeybee Params
        phx_window_type.u_value_glass = _hb_win_const.u_factor
        phx_window_type.glass_g_value = _hb_win_const.shgc

    # -- Frames ---------------------------------------------------------------
    if ph_params.ph_frame:
        # -- Use Detailed PH-Params
        phx_window_type.frame_type_display_name = ph_params.ph_frame.display_name
        phx_window_type.frame_top.u_value = ph_params.ph_frame.top.u_factor
        phx_window_type.frame_top.width = ph_params.ph_frame.top.width
        phx_window_type.frame_top.psi_glazing = ph_params.ph_frame.top.psi_glazing
        phx_window_type.frame_top.psi_install = ph_params.ph_frame.top.psi_install

        phx_window_type.frame_right.u_value = ph_params.ph_frame.right.u_factor
        phx_window_type.frame_right.width = ph_params.ph_frame.right.width
        phx_window_type.frame_right.psi_glazing = ph_params.ph_frame.right.psi_glazing
        phx_window_type.frame_right.psi_install = ph_params.ph_frame.right.psi_install

        phx_window_type.frame_bottom.u_value = ph_params.ph_frame.bottom.u_factor
        phx_window_type.frame_bottom.width = ph_params.ph_frame.bottom.width
        phx_window_type.frame_bottom.psi_glazing = ph_params.ph_frame.bottom.psi_glazing
        phx_window_type.frame_bottom.psi_install = ph_params.ph_frame.bottom.psi_install

        phx_window_type.frame_left.u_value = ph_params.ph_frame.left.u_factor
        phx_window_type.frame_left.width = ph_params.ph_frame.left.width
        phx_window_type.frame_left.psi_glazing = ph_params.ph_frame.left.psi_glazing
        phx_window_type.frame_left.psi_install = ph_params.ph_frame.left.psi_install
    else:
        # -- Use the basic Honeybee Params
        phx_window_type.frame_top.u_value = _hb_win_const.u_factor
        phx_window_type.frame_top.width = 0.1
        phx_window_type.frame_top.psi_glazing = 0.0
        phx_window_type.frame_top.psi_install = 0.0

        phx_window_type.frame_right.u_value = _hb_win_const.u_factor
        phx_window_type.frame_right.width = 0.1
        phx_window_type.frame_right.psi_glazing = 0.0
        phx_window_type.frame_right.psi_install = 0.0

        phx_window_type.frame_bottom.u_value = _hb_win_const.u_factor
        phx_window_type.frame_bottom.width = 0.1
        phx_window_type.frame_bottom.psi_glazing = 0.0
        phx_window_type.frame_bottom.psi_install = 0.0

        phx_window_type.frame_left.u_value = _hb_win_const.u_factor
        phx_window_type.frame_left.width = 0.1
        phx_window_type.frame_left.psi_glazing = 0.0
        phx_window_type.frame_left.psi_install = 0.0

    # -- Window Params as per ISO-10077-1 -------------------------------------
    if ph_params.ph_frame and ph_params.ph_glazing:
        phx_window_type.frame_factor = iso_10077_1.calculate_window_frame_factor(
            ph_params.ph_frame, ph_params.ph_glazing)
        phx_window_type.u_value_window = iso_10077_1.calculate_window_uw(
            ph_params.ph_frame, ph_params.ph_glazing)
    else:
        phx_window_type.frame_factor = 0.75
        phx_window_type.u_value_window = _hb_win_const.u_factor

    return phx_window_type


def build_transparent_assemblies_from_HB_Model(_project: project.PhxProject, _hb_model: model.Model) -> None:
    """Create PHX-WindowTypes (constructions) from an HB Model and add to the PHX-Project

    Will also align the id_nums of the Aperture Construction's with the WindowType in the Project dict.

    Arguments:
    ----------
        * _project (_project: project.PhxProject): The PhxProject to store the new window type on.
        * _hb_model (model.Model): The Honeybee Model to use as the source.

    Returns:
    --------
        * None
    """

    for room in _hb_model.rooms:
        for face in room.faces:
            for aperture in face._apertures:
                hb_win_construction = aperture.properties.energy.construction

                if hb_win_construction.identifier not in _project.window_types:

                    phx_aperture_constr = build_phx_window_type_from_hb_win_construction(
                        hb_win_construction)

                    _project.window_types[hb_win_construction.identifier] = phx_aperture_constr

                hb_win_construction.properties.ph.id_num = _project.window_types[
                    hb_win_construction.identifier].id_num

    return None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Areas" worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.xl_data import col_offset
from PHX.to_PHPP.phpp_model import areas_surface, areas_data, areas_thermal_bridges
from PHX.to_PHPP.phpp_localization import shape_model


class AreasInputLocation:
    """Generic input item for Areas worksheet items."""

    def __init__(self, _xl: xl_app.XLConnection, _sheet_name: str, _search_col: str, _search_item: str, _input_row_offset: int):
        self.xl = _xl
        self.sheet_name = _sheet_name
        self.search_col = _search_col
        self.search_item = _search_item
        self.input_row_offset = _input_row_offset

    def find_input_row(self, _row_start: int = 1, _row_end: int = 200) -> int:
        """Return the row number where the search-item is found input."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.sheet_name,
            _col=self.search_col,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.search_item in str(val):
                return i + self.input_row_offset

        raise Exception(
            f'\n\tError: Not able to find the "{self.search_item}" input '
            f'section of the "{self.sheet_name}" worksheet? Please be sure '
            f'the item is note with the "{self.search_item}" flag in column {self.search_col}?'
        )


class Surfaces:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Areas):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the 'Area input' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.surface_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if val == self.shape.surface_rows.locator_string_header:
                return i

        raise Exception(
            f'\n\tError: Not able to find the "Areas input" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section begins '
            f'with the "{self.shape.surface_rows.locator_string_header}" flag in '
            f'column {self.shape.surface_rows.locator_col_header}.'
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Area input' section."""

        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.surface_rows.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.surface_rows.locator_string_entry:
                return i

        raise Exception(
            f'\n\tError: Not able to find the first surface entry row in the "Areas input" section?'
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()

    def get_surface_phpp_id_by_name(self, _name: str) -> str:
        """Return the PHPP-Style id ("1-NorthRoofSurface", ...) when given the surface name."""

        if not self.section_first_entry_row:
            self.section_first_entry_row = self.find_section_header_row()

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=self.section_first_entry_row,
            row_end=self.section_first_entry_row+500,
            col=self.shape.surface_rows.inputs.description.column,
            find=_name
        )

        if not row:
            raise Exception(
                f'Error: Cannot locate the phpp surface named: {_name} in'
                f'column {self.shape.surface_rows.inputs.description.column}?'
            )

        prefix = self.xl.get_data(
            self.shape.name,
            f'{col_offset(self.shape.surface_rows.inputs.description.column, -1)}{row}'
        )

        print(f'Getting PHPP Surface id for {_name}')

        return f'{int(prefix)}-{_name}'


class ThermalBridges:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Areas):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None
    
    def find_section_header_row(self, _row_start: int = 100, _row_end: int = 500) -> int:
        """Return the row number of the 'Thermal Bridge input' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.thermal_bridge_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if val == self.shape.thermal_bridge_rows.locator_string_header:
                return i

        raise Exception(
            f'\n\tError: Not able to find the "Thermal Bridge input" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section begins '
            f'with the "{self.shape.thermal_bridge_rows.locator_string_header}" flag in '
            f'column {self.shape.thermal_bridge_rows.locator_col_header}.'
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Thermal Bridge input' section."""

        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.thermal_bridge_rows.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.thermal_bridge_rows.locator_string_entry:
                return i

        raise Exception(
            f'\n\tError: Not able to find the first Thermal Bridge entry row in the "Thermal Bridge input" section?'
        )


class Areas:
    """IO Controller for the PHPP Areas worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Areas):
        self.xl = _xl
        self.shape = _shape
        self.surfaces = Surfaces(self.xl, self.shape)
        self.thermal_bridges = ThermalBridges(self.xl, self.shape)

    def write_thermal_bridges(self, _tbs: List[areas_thermal_bridges.ThermalBridgeRow]) -> None:
        if not self.thermal_bridges.section_first_entry_row:
            self.thermal_bridges.section_first_entry_row = self.thermal_bridges.find_section_first_entry_row()
        
        for i, tb in enumerate(_tbs, start=self.thermal_bridges.section_first_entry_row):
            for item in tb.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_surfaces(self, _surfaces: List[areas_surface.SurfaceRow]) -> None:
        if not self.surfaces.section_first_entry_row:
            self.surfaces.section_first_entry_row = self.surfaces.find_section_first_entry_row()

        for i, surface in enumerate(_surfaces, start=self.surfaces.section_first_entry_row):
            for item in surface.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def _create_input_location_object(self, _phpp_model_obj: areas_data.AreasInput) -> AreasInputLocation:
        """Create and setup the AreasInputLocation object with the correct data."""
        phpp_obj_shape: shape_model.AreasDataInput = getattr(
            self.shape, _phpp_model_obj.input_type)
        return AreasInputLocation(
            _xl=self.xl,
            _sheet_name=self.shape.name,
            _search_col=phpp_obj_shape.locator_col,
            _search_item=phpp_obj_shape.locator_string,
            _input_row_offset=phpp_obj_shape.input_row_offset
        )

    def write_item(self, _phpp_model_obj: areas_data.AreasInput) -> None:
        """Write the VerificationInputItem item out to the PHPP Areas Worksheet."""
        input_object = self._create_input_location_object(_phpp_model_obj)
        input_row = input_object.find_input_row()
        xl_item = _phpp_model_obj.create_xl_item(self.shape.name, input_row)
        self.xl.write_xl_item(xl_item)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to convert a standard HBJSON Model over to WUFI Objects"""

from collections import defaultdict
from typing import Tuple, List

from honeybee import model
from honeybee import room

from PHX.model.project import PhxProject
from PHX.from_HBJSON import cleanup, create_assemblies, create_variant, create_shades
from PHX.from_HBJSON import create_schedules


class MissingPropertiesError(Exception):
    def __init__(self, _lbt_obj):
        self.message = (f'Error: LBT Object "{_lbt_obj}" does not have a .properties attribute?\n'
                        'Can not add the .ph to missing .properties attribute.')
        super().__init__(self.message)


def sort_hb_rooms_by_bldg_segment(_hb_rooms: Tuple[room.Room]) -> List[List[room.Room]]:
    """Returns Groups of Honeybee-Rooms broken up by properties.ph.ph_bldg_segment.identifier.

    Arguments:
    ----------
        * _hb_rooms (list[room.Room]): The list of Honeybee Rooms to sort into bins.

    Returns:
    --------
        * list[list[room.Room]]: A list of the groups of Honeybee Rooms.
    """

    rooms_by_segment = defaultdict(list)
    for room in _hb_rooms:
        rooms_by_segment[room.properties.ph.ph_bldg_segment.identifier].append(
            room)
    return list(rooms_by_segment.values())


def convert_hb_model_to_PhxProject(_hb_model: model.Model, group_components: bool = False) -> PhxProject:
    """Return a complete WUFI Project object with values based on the HB Model

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to base the WUFI Project on
        * group_components (bool): default=False. Set to true to have the converter
            group the components by assembly-type.

    Returns:
    --------
        * (Project): The new WUFI Project object. 
    """

    project = PhxProject()
    create_assemblies.build_opaque_assemblies_from_HB_model(project, _hb_model)
    create_assemblies.build_transparent_assemblies_from_HB_Model(project, _hb_model)
    create_schedules.build_util_patterns_ventilation_from_HB_Model(project, _hb_model)

    # -- TODO: Make all these operations if..else... with flags in the func arguments
    # -- Merge the rooms together by their Building Segment, Add to the Project
    # -- then create a new variant from the merged room.
    # -- try and weld the vertices too in order to reduce load-time.
    for room_group in sort_hb_rooms_by_bldg_segment(_hb_model.rooms):
        merged_hb_room = cleanup.merge_rooms(room_group)
        
        new_variant = create_variant.from_hb_room(
            merged_hb_room, project.assembly_types, project.window_types, group_components)
        
        new_variant = cleanup.weld_vertices(new_variant)
        
        create_shades.add_hb_model_shades_to_variant(new_variant, _hb_model)

        project.add_new_variant(new_variant)

    return project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to output a PHX Domain model to  a WUFI XML text document.

The PHX objects here are passed along to the xml_schemas for writing to file, 
and these xml-schemas contain the actual WUFI XML node names and ordering.
"""



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Fresh-Air Ventilation Classes"""

from __future__ import annotations
from typing import ClassVar
from dataclasses import dataclass, field


@dataclass
class PhxVentilationFlowrates:
    flow_supply: float = 0.0
    flow_extract: float = 0.0
    flow_transfer: float = 0.0


@dataclass
class PhxRoomVentilation:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    display_name: str = 'Unnamed_Space'
    wufi_type: int = 99  # User Determined
    quantity: int = 1
    floor_area: float = 0.0
    weighted_floor_area: float = 0.0
    net_volume: float = 0.0
    clear_height: float = 2.5

    # -- Ventilation related items
    vent_unit_id_num: int = 0
    vent_pattern_id_num: int = 0
    flow_rates: PhxVentilationFlowrates = field(default_factory=PhxVentilationFlowrates)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for Mech Equipment Options."""

from enum import Enum


class ComponentFaceType(Enum):
    WALL = 1
    FLOOR = 2
    ROOF_CEILING = 3
    AIR_BOUNDARY = 3
    WINDOW = 4


class ComponentExposureExterior(Enum):
    EXTERIOR = -1
    GROUND = -2
    SURFACE = -3


class ComponentFaceOpacity(Enum):
    OPAQUE = 1
    TRANSPARENT = 2
    AIRBOUNDARY = 3


class ComponentColor(Enum):
    EXT_WALL_INNER = 1
    EXT_WALL_OUTER = 2
    INNER_WALL = 3
    WINDOW = 4
    FLOOR = 5
    CEILING = 6
    SLOPED_ROOF_OUTER = 7
    SLOPED_ROOF_INNER = 8
    SLOPED_ROOF_THATCH = 9  # WTF?
    FLAT_ROOF_OUTER = 10
    FLAT_ROOF_INNER = 11
    SURFACE_GROUND_CONTACT = 12
    GROUND_ABOVE = 13
    GROUND_BENEATH = 14


class ThermalBridgeType(Enum):
    AMBIENT = 15
    PERIMETER = 16
    UNDERGROUND = 17


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create a new PhxBuilding from Honeybee-Rooms"""

from typing import List, Union, Dict

from honeybee import room, aperture, face

from PHX.model import building, constructions, components
from PHX.from_HBJSON import create_rooms, create_geometry
from PHX.model.enums.building import (ComponentExposureExterior, ComponentFaceOpacity,
                                ComponentColor, ComponentFaceType, ThermalBridgeType)


def _hb_face_opacity_to_phx_enum(_hb_face: face.Face) -> ComponentFaceOpacity:
    mapping = {
        "Wall": ComponentFaceOpacity.OPAQUE,
        "RoofCeiling": ComponentFaceOpacity.OPAQUE,
        "Floor": ComponentFaceOpacity.OPAQUE,
        "AirBoundary": ComponentFaceOpacity.AIRBOUNDARY,
    }
    return mapping[str(_hb_face.type)]


def _hb_face_type_to_phx_enum(_hb_face:  face.Face) -> ComponentFaceType:
    mapping = {
        'Wall': ComponentFaceType.WALL,
        'Floor': ComponentFaceType.FLOOR,
        'RoofCeiling': ComponentFaceType.ROOF_CEILING,
        'AirBoundary': ComponentFaceType.AIR_BOUNDARY,
    }
    return mapping[str(_hb_face.type)]


def _hb_ext_exposure_to_phx_enum(_hb_face: Union[face.Face, aperture.Aperture]) -> ComponentExposureExterior:
    mapping = {
        "Outdoors": ComponentExposureExterior.EXTERIOR,
        "Ground": ComponentExposureExterior.GROUND,
        "Surface": ComponentExposureExterior.SURFACE,
        "Adiabatic": ComponentExposureExterior.SURFACE
    }
    return mapping[str(_hb_face.boundary_condition)]


def _hb_int_color_to_phx_enum(_hb_face: face.Face) -> ComponentColor:
    mapping = {
        "Wall": {
            "Outdoors": ComponentColor.EXT_WALL_OUTER,
            "Surface": ComponentColor.INNER_WALL,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.INNER_WALL,
        },
        "RoofCeiling": {
            "Outdoors": ComponentColor.SLOPED_ROOF_OUTER,
            "Surface": ComponentColor.CEILING,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.CEILING,
        },
        "Floor": {
            "Outdoors": ComponentColor.FLOOR,
            "Surface": ComponentColor.FLOOR,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.FLOOR,
        },
    }
    return mapping[str(_hb_face.type)][str(_hb_face.boundary_condition)]


def _hb_ext_color_to_phx_enum(_hb_face: face.Face) -> ComponentColor:
    mapping = {
        "Wall": {
            "Outdoors": ComponentColor.EXT_WALL_OUTER,
            "Surface": ComponentColor.INNER_WALL,
            "Ground": ComponentColor.EXT_WALL_OUTER,
            "Adiabatic": ComponentColor.INNER_WALL,
        },
        "RoofCeiling": {
            "Outdoors": ComponentColor.SLOPED_ROOF_INNER,
            "Surface": ComponentColor.CEILING,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.CEILING,
        },
        "Floor": {
            "Outdoors": ComponentColor.FLOOR,
            "Surface": ComponentColor.FLOOR,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.FLOOR,
        },
    }
    return mapping[str(_hb_face.type)][str(_hb_face.boundary_condition)]


def create_component_from_hb_aperture(
        _host_compo: components.PhxComponentOpaque,
        _hb_aperture: aperture.Aperture,
        _hb_room: room.Room,
        _window_type_dict: Dict[str, constructions.PhxConstructionWindow]) -> building.PhxComponentAperture:
    """Create a new Transparent (window) Component based on a Honeybee Aperture.

    Arguments:
    ----------
        * _host (components.PhxComponentOpaque): The Aperture's parent opaque component.
        * _hb_aperture (aperture.Aperture): The Honeybee-Aperture to use as the source.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * componets.PhxComponentAperture: A new Transparent (window) Component.
    """
    
    # -- Create new Aperture
    phx_ap = components.PhxComponentAperture(_host=_host_compo)
    phx_ap.display_name = _hb_aperture.display_name
    phx_ap.exposure_interior = _hb_room.properties.ph.id_num
    phx_ap.window_type = _window_type_dict[_hb_aperture.properties.energy.construction.identifier]
    phx_ap.window_type_id_num = _hb_aperture.properties.energy.construction.properties.ph.id_num
    phx_ap.variant_type_name = _hb_aperture.properties.ph.variant_type

    # -- Create new Aperture Element (Sash)
    new_phx_ap_element = components.PhxApertureElement(_host=phx_ap)
    new_phx_ap_element.display_name = _hb_aperture.display_name
    new_phx_ap_element.polygon = create_geometry.create_PhxPolygonRectangular_from_hb_Face(_hb_aperture)
    
    if _hb_aperture.properties.ph.shading_dimensions:
        new_phx_ap_element.shading_dimensions.h_hori = _hb_aperture.properties.ph.shading_dimensions.h_hori
        new_phx_ap_element.shading_dimensions.d_hori = _hb_aperture.properties.ph.shading_dimensions.d_hori
        new_phx_ap_element.shading_dimensions.o_reveal = _hb_aperture.properties.ph.shading_dimensions.o_reveal
        new_phx_ap_element.shading_dimensions.d_reveal = _hb_aperture.properties.ph.shading_dimensions.d_reveal
        new_phx_ap_element.shading_dimensions.o_over = _hb_aperture.properties.ph.shading_dimensions.o_over
        new_phx_ap_element.shading_dimensions.d_over = _hb_aperture.properties.ph.shading_dimensions.d_over
    new_phx_ap_element.winter_shading_factor = _hb_aperture.properties.ph.winter_shading_factor
    new_phx_ap_element.summer_shading_factor = _hb_aperture.properties.ph.summer_shading_factor

    phx_ap.add_elements((new_phx_ap_element,))
    
    return phx_ap


def create_components_from_hb_face(_hb_face: face.Face,
                                   _hb_room: room.Room,
                                   _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
                                   _window_type_dict: Dict[str, constructions.PhxConstructionWindow]) -> components.PhxComponentOpaque:
    """Returns a new Opaque Component (and any child components) based on a Honeybee Face,

    Arguments:
    ----------
        * _hb_face (face.Face): The Honeybee-Face to use as the source.
        * _hb_room (room.Room)L The Honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * building.Component: The new Opaque Component.
    """

    opaque_compo = components.PhxComponentOpaque()
    opaque_compo.display_name = _hb_face.display_name
    opaque_compo.assembly = _assembly_dict[_hb_face.properties.energy.construction.identifier]
    opaque_compo.assembly_type_id_num = _hb_face.properties.energy.construction.properties.ph.id_num

    opaque_compo.face_type = _hb_face_type_to_phx_enum(_hb_face)
    opaque_compo.face_opacity = _hb_face_opacity_to_phx_enum(_hb_face)
    opaque_compo.exposure_exterior = _hb_ext_exposure_to_phx_enum(_hb_face)
    opaque_compo.exposure_interior = _hb_room.properties.ph.id_num
    opaque_compo.color_interior = _hb_int_color_to_phx_enum(_hb_face)
    opaque_compo.color_exterior = _hb_ext_color_to_phx_enum(_hb_face)

    # -- Create Polygon
    phx_polygon = create_geometry.create_PhxPolygon_from_hb_Face(_hb_face)
    opaque_compo.add_polygons(phx_polygon)

    # -- Create Child Apertures, register the Aperture with the Parent Compo
    for hb_aperture in _hb_face.apertures:
        phx_compo_aperture = create_component_from_hb_aperture(
            opaque_compo, hb_aperture, _hb_room, _window_type_dict)
        phx_polygon.add_child_poly_id(phx_compo_aperture.polygon_ids)
        opaque_compo.add_aperture(phx_compo_aperture)

    return opaque_compo


def create_components_from_hb_room(_hb_room: room.Room,
                                   _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
                                   _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
                                   ) -> List[components.PhxComponentOpaque]:
    """Create new Opaque and Transparent PHX-Components based on Honeybee-Room Faces.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _assembly_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * list[components.PhxComponentOpaque]: A list of the new opaque PhxComponents.
    """
    return [
        create_components_from_hb_face(
            hb_face, _hb_room, _assembly_dict, _window_type_dict)
        for hb_face in _hb_room
    ]


def create_zones_from_hb_room(_hb_room: room.Room) -> building.PhxZone:
    """Create a new PHX-Zone based on a honeybee-Room.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.

    Returns:
    --------
        * building.Zone: The new PHX-Zone object.
    """
    new_zone = building.PhxZone()

    new_zone.id_num = building.PhxZone._count
    new_zone.display_name = _hb_room.display_name

    # -- Sort the room order by full_name
    sorted_spaces = sorted(_hb_room.properties.ph.spaces,
                           key=lambda space: space.full_name)

    # -- Create a new WUFI-RoomVentilation for each space
    new_zone.wufi_rooms = [create_rooms.create_room_from_space(sp)
                           for sp in sorted_spaces]

    # -- Set Zone properties
    new_zone.volume_gross = _hb_room.volume
    new_zone.weighted_net_floor_area = sum(
        (rm.weighted_floor_area for rm in new_zone.wufi_rooms))
    new_zone.volume_net = sum((rm.net_volume for rm in new_zone.wufi_rooms))

    # Set the zone's occupancy based on the merged HB room
    new_zone.res_occupant_quantity = _hb_room.properties.energy.people.properties.ph.number_people
    new_zone.res_number_bedrooms = _hb_room.properties.energy.people.properties.ph.number_bedrooms

    return new_zone


def create_thermal_bridges_from_hb_room(_hb_room: room.Room) -> List[components.PhxComponentThermalBridge]:
    """Create a list of new PHX-ThermalBridges based on those found on a honeybee-Room.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.

    Returns:
    --------
        * (List[components.PhxThermalBridge]): A list of the new Thermal Bridge objects.
    """
    phx_thermal_bridges = []
    for thermal_bridge in sorted(_hb_room.properties.ph.ph_bldg_segment.thermal_bridges.values(), key=lambda tb:tb.display_name):
        phx_tb = components.PhxComponentThermalBridge()
        phx_tb.display_name = thermal_bridge.display_name
        phx_tb.quantity = thermal_bridge.quantity
        phx_tb.group_number = ThermalBridgeType(thermal_bridge.group_type.number)
        phx_tb.identifier = str(thermal_bridge.identifier)
        phx_tb.psi_value = thermal_bridge.psi_value
        phx_tb.fRsi_value = thermal_bridge.fRsi_value
        phx_tb.length = thermal_bridge.length
        phx_thermal_bridges.append(phx_tb)
    
    return phx_thermal_bridges


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Run script to convert an HBJSON file over to WUFI XML format."""

from os import mkdir
from typing import Tuple, List
import sys
import pathlib

from PHX.from_HBJSON import read_HBJSON_file, create_project
from PHX.to_WUFI_XML import xml_builder, xml_txt_to_file



class InputFileError(Exception):
    def __init__(self, path):
        self.msg = f'\nError: Cannot find HBJSON file: {path}'
        super().__init__(self.msg)


def resolve_paths(_args: List[str]) -> Tuple[pathlib.Path, pathlib.Path]:
    """Sort out the file input and output paths. Make the output directory if needed.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.

    Returns:
    --------
        * Tuple
            - [0] (pathlib.Path): The HBJSON Source file path.
            - [1] (pathlib.Path): The WUFI XML Target file path.
    """

    print('> Resolving file paths...')
    src = pathlib.Path(_args[1])
    if not src.exists():
        raise InputFileError(src)

    target_dir = pathlib.Path(_args[3])
    target_file = f'{_args[2]}.xml'

    if not target_dir.exists():
        mkdir(target_dir)

    target = pathlib.Path(target_dir, target_file)

    return src, target


if __name__ == '__main__':
    print("- " * 50)

    # --- Input / Output file Path
    # -----------------------------------------------------------------------------
    SOURCE_FILE, TARGET_FILE_XML = resolve_paths(sys.argv)

    # --- Read in the existing HB_JSON and re-build the HB Objects
    # -----------------------------------------------------------------------------
    print(f"> Reading in the HBJSON file: ./{SOURCE_FILE}")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(SOURCE_FILE)
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)

    # --- Generate the WUFI Project file.
    print(
        f'> Generating the PHX-Project from the Honeybee-Model: "{hb_model}"')
    phx_Project = create_project.convert_hb_model_to_PhxProject(
        hb_model, group_components=True)

    # --- Output the WUFI Project as an XML Text File
    # ---------------------------------------------------------------------------
    print(f'> Generating XML Text for the PHX-Project: "{phx_Project}"')
    xml_txt = xml_builder.generate_WUFI_XML_from_object(phx_Project)

    print(f"> Saving the XML file to: ./{TARGET_FILE_XML}")
    xml_txt_to_file.write_XML_text_file(TARGET_FILE_XML, xml_txt)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Geometry Classes"""

from __future__ import annotations
from dataclasses import dataclass, field
import math
from typing import ClassVar, Collection, List, Union, Optional


class PolygonEdgeError(Exception):
    def __init__(self, _polygon: PhxPolygonRectangular, _segment_name: str):
        self.msg = f"Error: Cannot create PhxPolygonRectangle {_polygon}"\
            f"segment {_segment_name}. Missing 1 or more vertices?"
        super().__init__(self.msg)


@dataclass
class PhxVertix:
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __eq__(self, other: PhxVertix) -> bool:
        return (self.x == other.x) and \
            (self.y == other.y) and \
            (self.z == other.z) and \
            (self.id_num == other.id_num)

    def __post_init__(self):
        PhxVertix._count += 1
        self.id_num = PhxVertix._count

    @property
    def unique_key(self) -> str:
        """Return a unique key (str) for the Vertex. Used for dicts, welding, etc"""
        return f"{self.x :0.10f}_{self.y :0.10f}_{self.z :0.10f}"

    def __hash__(self) -> int:
        return hash(self.unique_key)


@dataclass
class PhxVector:
    x: float
    y: float
    z: float

    @classmethod
    def from_2_points(cls, _start_pt: PhxVertix, _end_pt: PhxVertix) -> PhxVector:
        """Return a new PhxVector based on a start and end point."""
        return cls(
            _end_pt.x - _start_pt.x,
            _end_pt.y - _start_pt.y,
            _end_pt.z - _start_pt.z
        )

    def scale(self, _factor: float) -> None:
        self.x = self.x * _factor
        self.y = self.y * _factor
        self.z = self.z * _factor


@dataclass
class PhxPlane:
    normal: PhxVector
    origin: PhxVertix
    x: PhxVector
    y: PhxVector


@dataclass
class PhxLineSegment:
    vertix_1: PhxVertix
    vertix_2: PhxVertix

    @property
    def length(self) -> float:
        return abs(
            math.sqrt(
                (
                    (self.vertix_2.x - self.vertix_1.x)**2 +
                    (self.vertix_2.y - self.vertix_1.y)**2 +
                    (self.vertix_2.z - self.vertix_1.z)**2
                )
            )
        )


@dataclass
class PhxPolygon:
    """A Polygon surface defined by 3 or more vertices."""
    _count: ClassVar[int] = 0

    _display_name: str
    area: float
    center: PhxVertix
    normal_vector: PhxVector
    plane: PhxPlane

    _vertices: List[PhxVertix] = field(init=False, default_factory=list)
    child_polygon_ids: List[int] = field(init=False, default_factory=list)

    id_num: int = field(init=False, default=0)

    def __post_init__(self):
        PhxPolygon._count += 1
        self.id_num = PhxPolygon._count

    @property
    def display_name(self) -> str:
        if not self._display_name:
            return str(self.id_num)
        else:
            return self._display_name

    @display_name.setter
    def display_name(self, _in: str):
        self._display_name = str(_in)

    @property
    def vertices(self) -> List[PhxVertix]:
        return self._vertices

    @property
    def vertices_id_numbers(self) -> List[int]:
        return [v.id_num for v in self.vertices]

    @property
    def angle_from_horizontal(self) -> float:
        """Return the surface normal's angle (degrees) off horizontal.

        ie: pointing 'up'=0, vertical surface=90, pointing 'down'=180

        https://www.omnicalculator.com/math/angle-between-two-vectors

        Returns:
        -------
            * (float): Degrees. The surface normal angle off horizontal.
        """

        # Get the input Vector's X and Y parts
        a = self.normal_vector
        b = PhxVector(0, 0, 1.0)

        try:
            angle = math.acos(
                (a.x * b.x + a.y * b.y + a.z * b.z) /
                (math.sqrt(a.x**2 + a.y**2 + a.z**2) *
                 math.sqrt(b.x**2 + b.y**2 + b.z**2))
            )
        except ZeroDivisionError:
            angle = 0.0

        return math.degrees(angle)

    @property
    def is_horizontal(self, tolerance: float = 0.0001) -> bool:
        a = self.angle_from_horizontal
        if abs(a) < tolerance:
            return True
        elif abs(180 - abs(a)) < tolerance:
            return True
        return False

    @property
    def is_vertical(self, tolerance: float = 0.0001) -> bool:
        a = self.angle_from_horizontal
        if abs(90 - abs(a)) < tolerance:
            return True
        return False

    @property
    def cardinal_orientation_angle(self, _reference_vector: Optional[PhxVector] = None) -> float:
        """Calculate polygon normal's horizontal angle off a reference. By default, the
        reference vector will be (x=0,y=1,z=0) assuming north is y-direction.

        http://frasergreenroyd.com/obtaining-the-angle-between-two-vectors-for-360-degrees/
        Results 0=north, 90=east, 180=south, 270=west

        Arguments:
        ----------
            * _reference_vector: (PhxVector) The reference vector representing 
                'North'. By default, will use (x=0, y=1, z=0) as the north vector.

        Returns:
        --------
            (float) Degrees. The clockwise angle off reference of the surface normal.
        """
        if self.is_horizontal:
            return 0.0

        if not _reference_vector:
            _reference_vector = PhxVector(0, 1.0, 0)

        a_x = self.normal_vector.x
        a_y = self.normal_vector.y

        b_x = _reference_vector.x
        b_y = _reference_vector.y

        angle = math.atan2(b_y, b_x) - math.atan2(a_y, a_x)
        angle = math.degrees(angle)

        if angle < 0:
            angle = angle + 360

        return angle

    def add_vertix(self, _phx_vertix: PhxVertix) -> None:
        self.vertices.append(_phx_vertix)

    def add_child_poly_id(self, _child_ids: Union[Collection[int], int]) -> None:
        if not isinstance(_child_ids, Collection):
            _child_ids = (_child_ids,)

        for child_id in _child_ids:
            self.child_polygon_ids.append(child_id)


@dataclass
class PhxPolygonRectangular(PhxPolygon):
    """A Polygon with additional geometric attributes for rectangular surfaces."""

    vertix_upper_left: Optional[PhxVertix] = field(init=False, default=None)
    vertix_lower_left: Optional[PhxVertix] = field(init=False, default=None)
    vertix_lower_right: Optional[PhxVertix] = field(init=False, default=None)
    vertix_upper_right: Optional[PhxVertix] = field(init=False, default=None)

    def __post_init__(self):
        super().__post_init__()

    @property
    def edge_top(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Top' side of the Polygon (viewed from outside)."""
        if not self.vertix_upper_right or not self.vertix_upper_left:
            raise PolygonEdgeError(self, 'top')
        return PhxLineSegment(self.vertix_upper_right, self.vertix_upper_left)

    @property
    def edge_left(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Left' side of the Polygon (viewed from outside)."""
        if not self.vertix_upper_left or not self.vertix_lower_left:
            raise PolygonEdgeError(self, 'left')
        return PhxLineSegment(self.vertix_upper_left, self.vertix_lower_left)

    @property
    def edge_bottom(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Bottom' side of the Polygon (viewed from outside)."""
        if not self.vertix_lower_left or not self.vertix_lower_right:
            raise PolygonEdgeError(self, 'bottom')
        return PhxLineSegment(self.vertix_lower_left, self.vertix_lower_right)

    @property
    def edge_right(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Right' side of the Polygon (viewed from outside)."""
        if not self.vertix_lower_right or not self.vertix_upper_right:
            raise PolygonEdgeError(self, 'right')
        return PhxLineSegment(self.vertix_lower_right, self.vertix_upper_right)

    @property
    def width(self) -> float:
        return self.edge_top.length

    @property
    def height(self) -> float:
        return self.edge_left.length

    def add_vertix(self, _phx_vertix: PhxVertix) -> None:
        print(
            f'Method "add_vertix()" is not allowed for {self.__class__.__name__} objects.'
            f'Please assign the vertix corners directly (vertix_upper_left-left, vertix_lower_left-right, etc..)'
        )

    @property
    def vertices(self) -> List[PhxVertix]:
        """Return a List of the PhxPolygonRectangle Vertices (counter-clockwise from upper-left)."""
        corner_verts = [
            self.vertix_upper_left,
            self.vertix_lower_left,
            self.vertix_lower_right,
            self.vertix_upper_right,
        ]
        # filter out any 'None' vertices
        return [v for v in corner_verts if v]


@dataclass
class PhxGraphics3D:
    polygons: List[PhxPolygon] = field(default_factory=list)

    @property
    def vertices(self) -> List[PhxVertix]:
        """Returns a sorted list with all of the unique vertix objects of all the polygons in the collection."""
        return sorted(
            {vertix
             for polygon in self.polygons
             for vertix in polygon.vertices
             },
            key=lambda _: _.id_num)

    def add_polygons(self, _polygons: Union[Collection[PhxPolygon], PhxPolygon]) -> None:
        """Adds a new Polygon object to the collection"""

        if not isinstance(_polygons, Collection):
            _polygons = [_polygons]

        for polygon in _polygons:
            self.polygons.append(polygon)

    def get_polygons_by_id(self, _ids: Collection[int]) -> List[PhxPolygon]:
        """Returns a sorted list of polygons in the collection matching the IDs supplied.

        Arguments:
        ----------
            * _ids: (Collection[int]): A collection of one or more id_nums to look for.

        Returns:
        --------
            * List[PhxPolygon]: A sorted (by display_name) list of all the 
                polygons with matching id_nums.
        """

        if not isinstance(_ids, Collection):
            _ids = {_ids}

        return sorted(
            [p for p in self.polygons if p.id_num in _ids],
            key=lambda _: _.display_name
        )

    def __bool__(self) -> bool:
        return bool(self.polygons)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Areas / Thermal Bridges-Entry row"""

from dataclasses import dataclass
from typing import List, Tuple
from functools import partial

from PHX.model import components

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class ThermalBridgeRow:
    """A single Areas/Thermal Bridge entry row."""

    __slots__ = ('shape', 'phx_tb')
    shape: shape_model.Areas
    phx_tb: components.PhxComponentThermalBridge

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f'{getattr(self.shape.thermal_bridge_rows.inputs, _field_name).column}{_row_num}'
    
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.thermal_bridge_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Thermal Bridge Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAreas = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemAreas(create_range('description'), self.phx_tb.display_name),
            XLItemAreas(create_range('group_number'), self.phx_tb.group_number.value),
            XLItemAreas(create_range('quantity'), self.phx_tb.quantity),
            XLItemAreas(create_range('length'), self.phx_tb.length, "M", self._get_target_unit('length')),
            XLItemAreas(create_range('psi_value'), self.phx_tb.psi_value, "W/MK", self._get_target_unit('psi_value')),
            XLItemAreas(create_range('fRsi_value'), self.phx_tb.fRsi_value),
        ]


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create PHX-HVAC objects from Honeybee-Energy-PH HVAC"""

from typing import TypeVar, Union

from honeybee_energy_ph.hvac import ventilation, heating, cooling, _base
from PHX.model import hvac
from PHX.model.hvac.heating import AnyPhxHeater
from PHX.model.hvac.ventilation import AnyPhxVentilation
from PHX.model.hvac.cooling import AnyPhxCooling


T = TypeVar('T', bound=Union[AnyPhxVentilation, AnyPhxHeater, AnyPhxCooling]) 

def _transfer_attributes(_hbeph_obj: _base._PhHVACBase,
                         _phx_obj: T) -> T:
    """Copy the common attributes from a Honeybee-Energy-PH obj to a PHX-object

    Arguments:
    ---------
        * _hbeph_obj (_base._PhHVACBase): A Honeybee-Energy-PH HVAC Object.
        * _phx_obj (mech_equip.PhxMechanicalEquipment): A PHX-Mechanical
            Equipment object.

    Returns:
    --------
        * (mech_equip.PhxMechanicalEquipment): The input PHX Mechanical Equipment 
            object, with its attribute values set to match the Honeybee-PH object.
    """

    # -- Copy the base scope attributes from HB->PHX
    hb_attrs = {attr_name for attr_name in dir(
        _hbeph_obj) if not attr_name.startswith('_')}
    phx_base_attrs = {attr_name for attr_name in dir(
        _phx_obj) if not attr_name.startswith('_')}

    for attr_nm in hb_attrs.intersection(phx_base_attrs):
        setattr(_phx_obj, attr_nm, getattr(_hbeph_obj, attr_nm))

    # -- Also copy attrs to PHX.params for objects which have them
    if _phx_obj.params:
        phx_params_attrs = {attr_name for attr_name in dir(
            _phx_obj.params) if not attr_name.startswith('_')}
        for attr_nm in hb_attrs.intersection(phx_params_attrs):
            setattr(_phx_obj.params, attr_nm, getattr(_hbeph_obj, attr_nm))

    return _phx_obj


# -----------------------------------------------------------------------------
# -- Ventilation

def _default_ventilator_name(_hr:float, _mr:float) -> str:
    """Return a Ventilation Unit name
    
    Arguments:
    ----------
        * _hr (float): The Heat Recovery efficiency of the unit.    
        * _mr (float): The Moisture Recovery efficiency of the unit.    
    Returns:
    --------
        * (str)
    """
    return f"Ventilator: {_hr*100 :0.0f}%-HR, {_mr*100 :0.0f}%-MR"

def build_phx_ventilator(_hbeph_vent_sys: ventilation.PhVentilationSystem) -> hvac.PhxDeviceVentilator:
    """Returns a new Fresh-Air Ventilator built from the hb-energy hvac parameters.

    This will look at the Space's Host-Room .properties.energy.hvac for data.

    Arguments:
    ----------
        *_hbeph_vent (ventilation.PhVentilationSystem): The Honeybee-PH Ventilation System
            to build the PHX-Ventilation from.

    Returns:
    --------
        * (mech_equip.Ventilator): The new Passive House Ventilator created.
    """

    phx_vent = hvac.PhxDeviceVentilator()
    _id_num = phx_vent.id_num # preserver so 'transfer_attributes' doesn't kill it...

    if not _hbeph_vent_sys.ventilation_unit:
        return phx_vent
    phx_vent = _transfer_attributes(_hbeph_vent_sys, phx_vent)
    phx_vent = _transfer_attributes(_hbeph_vent_sys.ventilation_unit, phx_vent)

    # -- Sort out the Display name to use
    if not _hbeph_vent_sys.ventilation_unit:
        ventilator_name = _default_ventilator_name(0.0, 0.0)
    elif _hbeph_vent_sys.ventilation_unit.display_name == "_unnamed_ventilator_":
        ventilator_name = _default_ventilator_name(
            _hbeph_vent_sys.ventilation_unit.sensible_heat_recovery,
            _hbeph_vent_sys.ventilation_unit.latent_heat_recovery
        )
    else:
        ventilator_name = _hbeph_vent_sys.ventilation_unit.display_name

    phx_vent.display_name = ventilator_name
    phx_vent.id_num = _id_num

    return phx_vent


def build_phx_ventilation_sys(_hbeph_vent: ventilation.PhVentilationSystem) -> hvac.PhxDeviceVentilator:
    """Build a new PHX Ventilation Mechanical Device.

    Arguments:
    ----------
        *_hbeph_vent (ventilation.PhVentilationSystem): The Honeybee-PH Ventilation System
            to build the PHX-Ventilation from.

    Returns:
    --------
        * (mech.PhxDeviceVentilator): A new mech ventilation device.
    """

    phx_ventilator = build_phx_ventilator(_hbeph_vent)

    # TODO: Distribution...

    return phx_ventilator


# -----------------------------------------------------------------------------
# -- Heating


def build_phx_heating_electric(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterElectric:
    phx_obj = hvac.PhxHeaterElectric()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_fossil_boiler(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterBoilerFossil:
    phx_obj = hvac.PhxHeaterBoilerFossil()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_wood_boiler(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterBoilerWood:
    phx_obj = hvac.PhxHeaterBoilerWood()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_district(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterDistrictHeat:
    phx_obj = hvac.PhxHeaterDistrictHeat()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_annual(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterHeatPumpAnnual:
    phx_obj = hvac.PhxHeaterHeatPumpAnnual()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_monthly(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterHeatPumpMonthly:
    phx_obj = hvac.PhxHeaterHeatPumpMonthly()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_combined(_hbeph_heater: heating.PhHeatingSystem) -> hvac.PhxHeaterHeatPumpCombined:
    phx_obj = hvac.PhxHeaterHeatPumpCombined()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_device(_htg_sys: heating.PhHeatingSystem) -> hvac.PhxHeatingDevice:
    """Returns a new PHX-Heating-Device based on an input Honeybee-PH Heating System.

    Arguments:
    ----------
        * _htg_sys (heating.PhHeatingSystem): The Honeybee-PH Heating System to build 
            the new PHX Heating system from.

    Returns:
    --------
        * (mech.heating.PhxHeatingDevice): The new PHX Heating System created.
    """

    # -- Mapping Honeybee-PH -> PHX types
    phx_heater_classes = {
        'PhHeatingDirectElectric': build_phx_heating_electric,
        'PhHeatingFossilBoiler': build_phx_heating_fossil_boiler,
        'PhHeatingWoodBoiler': build_phx_heating_wood_boiler,
        'PhHeatingDistrict': build_phx_heating_district,
        'PhHeatingHeatPumpAnnual': build_phx_heating_hp_annual,
        'PhHeatingHeatPumpRatedMonthly': build_phx_heating_hp_monthly,
        'PhHeatingHeatPumpCombined': build_phx_heating_hp_combined,
    }

    # -- Get and build the right heater equipment type
    phx_heater = phx_heater_classes[_htg_sys.heating_type](_htg_sys)
    return phx_heater


def build_phx_heating_sys(_htg_sys: heating.PhHeatingSystem) -> hvac.PhxHeatingDevice:
    """

    Arguments:
    ----------
        * _htg_sys (heating.PhHeatingSystem): The Honeybee-PH Heating System to build 
            the new PHX Heating system from.

    Returns:
    --------
        * (mech.PhxHeatingDevice): 
    """
    
    phx_htg_device = build_phx_heating_device(_htg_sys)

    # TODO: Distribution...

    return phx_htg_device


# -----------------------------------------------------------------------------
# --- Cooling


def build_phx_cooling_ventilation(_hbeph_cooling: cooling.PhCoolingSystem) -> hvac.PhxCoolingVentilation:
    phx_obj = hvac.PhxCoolingVentilation()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_recirculation(_hbeph_cooling: cooling.PhCoolingSystem) -> hvac.PhxCoolingRecirculation:
    phx_obj = hvac.PhxCoolingRecirculation()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_dehumidification(_hbeph_cooling: cooling.PhCoolingSystem) -> hvac.PhxCoolingDehumidification:
    phx_obj = hvac.PhxCoolingDehumidification()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_panel(_hbeph_cooling: cooling.PhCoolingSystem) -> hvac.PhxCoolingPanel:
    phx_obj = hvac.PhxCoolingPanel()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_device(_clg_sys: cooling.PhCoolingSystem) -> hvac.PhxCoolingDevice:
    """Returns a new PHX-Cooling-Device based on an input Honeybee-PH cooling System.

    Arguments:
    ----------
        * _clg_sys (cooling.PhCoolingSystem): The Honeybee-PH Cooling-System to build 
            the new PHX-Cooling-System from.

    Returns:
    --------
        * (mech.cooling.PhxHeatingDevice): The new PHX cooling System created.
    """

    # -- Mapping Honeybee-PH -> PHX types
    phx_cooling_classes = {
        'PhCoolingVentilation': build_phx_cooling_ventilation,
        'PhCoolingRecirculation': build_phx_cooling_recirculation,
        'PhCoolingDehumidification': build_phx_cooling_dehumidification,
        'PhCoolingPanel': build_phx_cooling_panel,
    }

    # -- Get and build the right heater equipment type
    phx_cooling = phx_cooling_classes[_clg_sys.cooling_class_name](_clg_sys)
    return phx_cooling


def build_phx_cooling_sys(_htg_sys: cooling.PhCoolingSystem) -> hvac.PhxCoolingDevice:
    """

    Arguments:
    ----------
        * _htg_sys (cooling.PhCoolingSystem): The Honeybee-PH cooling System to build 
            the new PHX cooling system from.

    Returns:
    --------
        * (mech.PhxCoolingDevice):
    """

    phx_htg_device = build_phx_cooling_device(_htg_sys)

    # TODO: Distribution...

    return phx_htg_device



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for Mech Equipment Options."""

from enum import Enum
from honeybee_energy_ph.hvac.hot_water import PhSHWTankType

class FuelType(Enum):
    GAS = 1
    OIL = 2
    WOOD_LOG = 3
    WOOD_PELLET = 4
    # HARD_COAL_CGS_70_PHC = 'HARD_COAL_CGS_70_PHC'
    # HARD_COAL_CGS_35_PHC = 'HARD_COAL_CGS_35_PHC'
    # HARD_COAL_HS_0_PHC = 'HARD_COAL_HS_0_PHC'
    # GAS_CGS_70_PHC = 'GAS_CGS_70_PHC'
    # GAS_CGS_35_PHC = 'GAS_CGS_35_PHC'
    # GAS_HS_0_PHC = 'GAS_HS_0_PHC'
    # OIL_CGS_70_PHC = 'OIL_CGS_70_PHC'
    # OIL_CGS_35_PHC = 'OIL_CGS_35_PHC'
    # OIL_HS_0_PHC = 'OIL_HS_0_PHC'


class SystemType(Enum):
    ANY = 0
    VENTILATION = 1
    ELECTRIC = 2
    BOILER = 3
    DISTRICT_HEAT = 4
    HEAT_PUMP = 5
    WATER_STORAGE = 8


class DeviceType(Enum):
    VENTILATION = 1
    ELECTRIC = 2
    BOILER = 3
    DISTRICT_HEAT = 4
    HEAT_PUMP = 5
    WATER_STORAGE = 8


class HeatPumpType(Enum):
    COMBINED = 2
    ANNUAL = 3
    RATED_MONTHLY = 4
    HOT_WATER = 5


class CoolingType(Enum):
    VENTILATION = 1
    RECIRCULATION = 2
    DEHUMIDIFICATION = 3
    PANEL = 4


class PhxHotWaterInputOptions(Enum):
    SPEC_TOTAL_LOSSES = 1
    SPEC_STANDBY_LOSSES = 2
    TOTAL_LOSSES = 3


class PhxHotWaterTankType(Enum):
    NONE = 0
    DHW_AND_HEATING = 1
    DHW_ONLY = 2

    @classmethod
    def from_hbph_type(cls, _hbph_type):
        """Return a new PhxHotWaterTankType based on the Honeybee-PH Tank Type."""
        _mapping = {
            "0-NO STORAGE TANK": cls.NONE,
            "1-DHW AND HEATING": cls.DHW_AND_HEATING,
            "2-DHW ONLY": cls.DHW_ONLY,
        }
        return _mapping[_hbph_type]


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for PHI Certification Settings (PHPP v9)."""

from enum import Enum

class PhiCertBuildingCategoryType(Enum):
    RESIDENTIAL_BUILDING = 1
    NONRESIDENTIAL_BUILDING = 2

class PhiCertBuildingUseType(Enum):
        DWELLING = 10
        NURSING_HOME = 11
        OTHER_RES = 12
        OFFICE = 20
        SCHOOL = 21
        OTHER_NONRES = 22

class PhiCertIHGType(Enum):
        STANDARD = 2
        RES_CUSTOM = 3
        NONRES_CUSTOM = 4

class PhiCertOccupancyType(Enum):
        STANDARD = 1
        CUSTOM = 2

class PhiCertType(Enum):
    PASSIVE_HOUSE = 1
    ENERPHIT = 2
    LOW_ENERGY_BUILDING = 3
    OTHER = 4


class PhiCertClass(Enum):
    CLASSIC = 1
    PLUS = 2
    PREMIUM = 3


class PhiCertificationPEType(Enum):
    PE = 1
    PER = 2


class PhiCertEnerPHitType(Enum):
    BY_COMPONENT = 1
    BY_DEMAND = 2


class PhiCertRetrofitType(Enum):
    NEW_BUILDING = 1
    RETROFIT = 2
    STEP_BY_STEP_RETROFIT = 3



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP 'Components' worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.xl_data import col_offset
from PHX.to_PHPP.phpp_model import component_frame, component_glazing, component_vent
from PHX.to_PHPP.phpp_localization import shape_model


class Glazings:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.glazings.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.glazings.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.glazings.locator_string_header}'"
            f"header on the '{self.shape.name}' sheet, column {self.shape.glazings.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.glazings.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.glazings.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.glazings.locator_string_entry}' "
            f" entry start on the '{self.shape.name}' sheet, column {self.shape.glazings.locator_col_entry}?"
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()

    def get_glazing_phpp_id_by_name(self, _name: str) -> Optional[str]:
        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=1,
            row_end=500,
            col=self.shape.glazings.inputs.description.column,
            find=_name)

        if not row:
            return

        prefix = self.xl.get_data(
            self.shape.name,
            f'{col_offset(self.shape.glazings.inputs.description.column, -1)}{row}'
        )

        print(f'Getting PHPP Glazing id for {_name}')

        return f'{prefix}-{_name}'


class Frames:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.frames.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.frames.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.frames.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.frames.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.frames.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.frames.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.frames.locator_string_entry}'"
            f"entry start on the 'Components' sheet, column {self.shape.frames.locator_col_entry}?"
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()

    def get_frame_phpp_id_by_name(self, _name: str, _row_start: int = 1, _row_end: int = 500) -> str:
        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=self.shape.frames.inputs.description.column,
            find=_name
        )

        if not row:
            msg = f'Error: Cannot find a Frame component named: "{_name}" in'\
                  f'column {self.shape.frames.inputs.description.column}?'
            raise Exception(msg)

        prefix = self.xl.get_data(
            self.shape.name,
            f'{col_offset(self.shape.frames.inputs.description.column, -1)}{row}'
        )

        print(f'Getting PHPP Frame id for {_name}')

        return f'{prefix}-{_name}'


class Ventilators:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilators.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.ventilators.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.ventilators.locator_string_header}' header on the "
            f"'{self.shape.name}' sheet, column {self.shape.ventilators.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilators.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.ventilators.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.ventilators.locator_string_entry}' entry start on "
            f"the '{self.shape.name}' sheet, column {self.shape.ventilators.locator_col_entry}?"
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()

    def get_ventilator_phpp_id_by_name(self, _name: str, _row_start: int = 1, _row_end: int = 500) -> str:
        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=self.shape.ventilators.inputs.display_name.column,
            find=_name
        )

        if not row:
            raise Exception(
                f'Error: Cannot find a Ventilator component named: "{_name}"]'
                f'in column {self.shape.ventilators.inputs.display_name.column}?'
            )

        prefix = self.xl.get_data(
            self.shape.name,
            f'{col_offset(self.shape.ventilators.inputs.display_name.column, -1)}{row}'
        )
        return f'{prefix}-{_name}'


class Components:
    """IO Controller for PHPP "Components" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Components):
        self.xl = _xl
        self.shape = shape
        self.glazings = Glazings(self.xl, self.shape)
        self.frames = Frames(self.xl, self.shape)
        self.ventilators = Ventilators(self.xl, self.shape)

    def write_glazings(self, _glazing_rows: List[component_glazing.GlazingRow]) -> None:
        if not self.glazings.section_first_entry_row:
            self.glazings.section_first_entry_row = self.glazings.find_section_first_entry_row()

        for i, glazing_row in enumerate(_glazing_rows, start=self.glazings.section_first_entry_row):
            for item in glazing_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_frames(self, _frame_row: List[component_frame.FrameRow]) -> None:
        if not self.frames.section_first_entry_row:
            self.frames.section_first_entry_row = self.frames.find_section_first_entry_row()

        for i, frame_row in enumerate(_frame_row, start=self.frames.section_first_entry_row):
            for item in frame_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_ventilators(self, _ventilator_row: List[component_vent.VentilatorRow]) -> None:
        if not self.ventilators.section_first_entry_row:
            self.ventilators.section_first_entry_row = self.ventilators.find_section_first_entry_row()

        for i, ventilator_row in enumerate(_ventilator_row, start=self.ventilators.section_first_entry_row):
            for item in ventilator_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)
